############################################################################# 
## 
#W  gpd.gi                    GAP4 package `Gpd'                Chris Wensley 
#W                                                               & Emma Moore
##
#H  @(#)Id: "gpd/gap/gpd.gi", v.1.05 2008/11/21 
## 
##  This file contains generic methods for groupoids. 
##
Revision.("gpd/gap/gpd_gi") :=
             "@(#)$Id: gpd.gi,v 1.05 2008/11/21 gap Exp $";

#############################################################################
##  Standard error messages

GPD_CONSTRUCTORS := Concatenation( 
    "The standard operations which construct a groupoid are:\n",
    "1.  SinglePieceGroupoid( object group, list of objects );\n",
    "2.  DomainWithSingleObject( group, single object );\n",
    "3.  UnionOfPieces( list of groupoids );\n",
    "4.  SinglePieceGroupoidByGenerators( list of elements );\n", 
    "5.  SinglePieceGroupoidWithRays( parent gpd, root gp, ray mults. );\n", 
    "6.  Groupoid( one of the previous parameter options );" );

##  these are called by the GlobalFunction Groupoid 

SUB_CONSTRUCTORS := Concatenation( 
    "The standard operations which construct a subgroupoid are:\n", 
    "1.  SubgroupoidByPieces( groupoid, list of [imobs,hom] pairs );\n",
    "2.  FullSubgroupoid( groupoid, list of objects );\n", 
    "3.  MaximalDiscreteSubgroupoid( groupoid );\n", 
    "4.  DiscreteSubgroupoid( groupoid, list of obs, list of subgps );\n",
    "5.  FullIdentitySubgroupoid( groupoid );\n", 
    "6.  DiscreteIdentitySubgroupoid( groupoid );\n", 
    "7.  Subgroupoid( one of the previous parameter options );" );

##  and these are called by the GlobalFunction Subgroupoid 

#############################################################################
##
#M  SinglePieceGroupoidNC                                            
#M  SinglePieceGroupoid                                            
##
InstallMethod( SinglePieceGroupoidNC, "method for a connected groupoid",
    true, [ IsGroup, IsHomogeneousList ], 0,
function( gp, obs ) 

    local  fam, filter, cf, gpd, fgpd;

    fam := CollectionsFamily( FamilyObj( gp ) );
    filter := IsMWOSinglePieceRep and IsMagmaWithObjectsAndInverses;
    cf := CollectionsFamily( FamilyObj( gp ) ); 
    gpd := Objectify( NewType( fam, filter ), 
           rec( objects := obs, magma := gp, eltsfam := cf ) ); 
    SetIsAssociative( gpd, true );
    SetIsSinglePieceDomain( gpd, true );
    SetIsAssociative( gpd, true ); 
    SetIsCommutative( gpd, IsCommutative( gp ) ); 
    #?  (08/09/08)  added the "else"
    if ( Length( obs ) = 1 ) then 
        SetIsDiscreteDomain( gpd, true );
    fi; 
    SetIsDirectProductWithCompleteGraphDomain( gpd, true ); 
    fgpd := FamilyObj( gpd ); 
    fgpd!.mwo := gpd; 
    return gpd; 
end );

InstallMethod( SinglePieceGroupoid, "method for a connected groupoid",
    true, [ IsGroup, IsHomogeneousList ], 0,
function( gp, obs ) 
        if not HasIsSSortedList( obs ) then 
            Sort( obs ); 
        fi; 
    if not IsDuplicateFree( obs ) then
        Error( "objects must be distinct" );
    fi; 
    return SinglePieceGroupoidNC( gp, obs );
end ); 

InstallMethod( SinglePieceGroupoidWithRaysNC, 
    "generic method for a connected gpd with variable object gps", true, 
    [ IsGroupoid, IsGroup, IsHomogeneousList ], 0,
function( pgpd, rgp, rays )

    local  obs, rob, fam, filter, gpd, id;

    fam := FamilyObj( [ pgpd, rgp, rays ] );
    filter := IsSinglePieceRaysRep;
    gpd := Objectify( NewType( fam, filter ), rec () ); 
    gpd!.objects := pgpd!.objects; 
    gpd!.magma := rgp; 
    gpd!.eltsfam := pgpd!.eltsfam; 
    gpd!.rayConjugators := rays;
    SetIsSinglePieceDomain( gpd, true ); 
    SetParent( gpd, pgpd ); 
    id := One( pgpd!.magma ); 
    SetIsDirectProductWithCompleteGraphDomain( gpd, 
        ForAll( rays, r -> r = id ) );
    return gpd; 
end );

InstallMethod( SinglePieceGroupoidWithRays, 
    "generic method for a connected gpd with variable object gps", true, 
    [ IsGroupoid and IsSinglePiece, IsGroup, IsHomogeneousList ], 0,
function( gpd, rgp, rays )

    local  obs, gp, par;

    if ( HasIsDirectProductWithCompleteGraph( gpd ) 
         and IsDirectProductWithCompleteGraph( gpd ) ) then 
        par := gpd; 
    else ## assume RaysRep 
        par := Parent( gpd ); 
    fi; 
    obs := par!.objects; 
    if not ( Length( obs ) = Length( rays ) ) then 
        Error( "should be one ray for each object in the parent" ); 
    fi; 
    gp := par!.magma; 
    if not IsSubgroup( gp, rgp ) then
        Error( "root group not a subgroup of the parent group" );
    fi;
    if not ForAll( rays, r -> r in gp ) then
        Error( "not all the rays are elements of the parent group" );
    fi; 
    if not ( rays[1] = One( gp ) ) then
        Error( "root group conjugator not the identity element" );
    fi;
    return SinglePieceGroupoidWithRaysNC( par, rgp, rays );
end );

#############################################################################
##
#M  SinglePieceGroupoidByGenerators
##
InstallMethod( SinglePieceGroupoidByGenerators, "for a list of elements",
    true, [ IsHomogeneousList ], 0,
function( gens ) 
    Print( "SinglePieceGroupoidByGenerators not yet implemented!\n" ); 
    return fail;
end );

#############################################################################
##
#M  PiecesOfGroupoid
##
##  InstallMethod( PiecesOfGroupoid, "for a connected groupoid",
##      true, [ IsGroupoid and IsSinglePiece ], 0,
##  function( G )
##      return [ G ];
##  end );

#############################################################################
##
#M  GeneratorsOfGroupoid
##
InstallMethod( GeneratorsOfGroupoid, "for a groupoid", true, [ IsGroupoid ], 
    0, GeneratorsOfMagmaWithObjects );

#############################################################################
##
#M  IsSinglePieceDomain
##
InstallMethod( IsSinglePieceDomain, "for a groupoid", true,
    [ IsGroupoid ], 0,
function( gpd )
    return ( Length( ObjectList( gpd ) ) = 1 );
end );

#############################################################################
##
#M  IsPermGroupoid
#M  IsFpGroupoid
#M  IsPcGroupoid
##
InstallMethod( IsPermGroupoid, "for a groupoid", true,
    [ IsGroupoid ], 0,
function( gpd ) 
    if IsSinglePiece( gpd ) then 
        return ( IsPermCollection(gpd!.magma) and IsPermGroup(gpd!.magma) ); 
    else 
        return ForAll( Pieces(gpd), c -> IsPermGroupoid(c) ); 
    fi; 
end );

InstallMethod( IsFpGroupoid, "for a groupoid", true,
    [ IsGroupoid ], 0,
function( gpd )
    if IsSinglePiece( gpd ) then 
        return ( IsGroupOfFamily(gpd!.magma) and IsFpGroup(gpd!.magma) ); 
    else 
        return ForAll( Pieces(gpd), c -> IsFpGroupoid(c) );   
    fi; 
end );

InstallMethod( IsPcGroupoid, "for a groupoid", true,
    [ IsGroupoid ], 0,
function( gpd )
    if IsSinglePiece( gpd ) then 
        return ( HasIsPolycyclicGroup( gpd!.magma )
                 and IsPolycyclicGroup( gpd!.magma ) ); 
    else 
        return ForAll( Pieces(gpd), c -> IsPcGroupoid(c) );   
    fi; 
end ); 

#############################################################################
##
#M  DomainWithSingleObject
##

InstallMethod( DomainWithSingleObject, "generic method for a group",
    true, [ IsGroup, IsObject ], 0,
function( gp, obj ) 
    local  o; 
    if ( IsList( obj ) and ( Length(obj) = 1 ) ) then
        o := obj[1]; 
    elif IsScalar( obj ) then 
        o := obj;
    else 
        Error( "usage: DomainWithSingleObject( <group>, <object> );" ); 
    fi; 
    return SinglePieceGroupoidNC( gp, [ o ] );
end );

#############################################################################
##
#F  Groupoid( <pieces> )                groupoid as list of pieces 
#F  Groupoid( <gp>, <obj> )             group as groupoid 
#F  Groupoid( <gp>, <obs> )             single piece groupoid 
#F  Groupoid( <gpd>, <rgp>, <rays> )    subgroupoid by root groups, rays 
##
InstallGlobalFunction( Groupoid, function( arg )

    local  nargs, id, rays;

    nargs := Length( arg ); 
    # list of pieces
    if ( ( nargs = 1 ) and IsList( arg[1] ) 
         and  ForAll( arg[1], G -> IsGroupoid(G) ) ) then
        Info( InfoGpd, 2, "ByUnion" );
        return UnionOfPieces( arg[1] );
    # group * tree groupoid
    elif ( ( nargs = 2 ) and IsList( arg[2] ) and IsGroup( arg[1] ) ) then
        Info( InfoGpd, 2, "group plus objects" ); 
        return SinglePieceGroupoid( arg[1], arg[2] );
    # one-object groupoid
    elif ( ( nargs = 2 ) and IsScalar( arg[2] ) and IsGroup( arg[1] ) ) then
        Info( InfoGpd, 2, "SingleObject" );
        return DomainWithSingleObject( arg[1], arg[2] );
    elif ( ( nargs = 3 ) and IsGroupoid( arg[1] ) and IsGroup( arg[2] ) 
           and IsHomogeneousList( arg[3] ) ) then 
        return SinglePieceGroupoidWithRays( arg[1], arg[2], arg[3] );
    else
        Info( InfoGpd, 1, GPD_CONSTRUCTORS );
        return fail;
    fi;
end );

#############################################################################
##
##  ViewObj
#M  PrintObj
##
InstallMethod( ViewObj, "for a groupoid", true, [ IsGroupoid ], 0, PrintObj ); 

InstallMethod( PrintObj, "for a groupoid", true, [ IsGroupoid ], 0,
function ( gpd )
    local comp, c, i;

    if ( HasIsPermGroupoid( gpd ) and IsPermGroupoid( gpd ) ) then
        Print( "Perm " );
    elif ( HasIsFpGroupoid( gpd ) and IsFpGroupoid( gpd ) ) then
        Print( "Fp " );
    elif ( HasIsPcGroupoid( gpd ) and IsPcGroupoid( gpd ) ) then
        Print( "Pc " );
    fi; 
    if IsSinglePiece( gpd ) then  
        if IsDirectProductWithCompleteGraph( gpd ) then 
            Print( "single piece groupoid: < " ); 
            Print( gpd!.magma, ", ", gpd!.objects, " >" );
        else 
            Print( "single piece groupoid with rays: < " ); 
            Print( gpd!.magma, ", ", gpd!.objects, ", ", 
                                     gpd!.rayConjugators, " >" );
        fi; 
    else 
        comp := Pieces( gpd );
        Print( "groupoid with pieces:\n" );
        for i in [1..Length(comp)] do
            c := comp[i]; 
            Print( i, ":  ", c, "\n" );
        od;
    fi;
end );

##############################################################################
##
#M  Display( <gpd> ) . . . . . . . . . . . . . . . . . . .  display a groupoid
##
InstallMethod( Display, "for a groupoid", [ IsGroupoid ],
function( gpd )
    
    local  comp, c, i, pgpd, gp, len, rgp, rays;

    if ( HasIsPermGroupoid( gpd ) and IsPermGroupoid( gpd ) ) then
        Print( "Perm " );
    elif ( HasIsFpGroupoid( gpd ) and IsFpGroupoid( gpd ) ) then
        Print( "Fp " );
    elif ( HasIsPcGroupoid( gpd ) and IsPcGroupoid( gpd ) ) then
        Print( "Pc " );
    fi;
    if IsSinglePiece( gpd ) then 
        if IsDirectProductWithCompleteGraph( gpd ) then 
            Print( "single piece groupoid: " );
            if HasName( gpd ) then
                Print( gpd );
            fi;
            Print( "\n" ); 
            Print( "  objects: ", gpd!.objects, "\n" );
            gp := gpd!.magma;
            Print( "    group: " );
            if HasName( gp ) then
                Print( gp, " = <", GeneratorsOfGroup( gp ), ">\n" );
            else
                Print( gp, "\n" );
            fi;
        else
            Print( "single piece groupoid with rays: " );
            if HasName( gpd ) then
                Print( gpd );
            fi;
            Print( "\n" ); 
            pgpd := Parent( gpd );
            Print( "parent groupoid: ", pgpd, "\n" ); 
            Print( "  objects: ", gpd!.objects, "\n" );
            rgp := gpd!.magma;
            Print( "  root group: " );
            if HasName( rgp ) then
                Print( rgp, " = <", GeneratorsOfGroup( rgp ), ">\n" );
            else
                Print( rgp, "\n" );
            fi; 
            Print( " conjugators: ", gpd!.rayConjugators, "\n" );
        fi; 
    else
        comp := Pieces( gpd );
        len := Length( comp );
        Print( "groupoid with ", len, " pieces:\n" );
        for i in [1..len] do
            c := comp[i];
            if IsDirectProductWithCompleteGraph( c ) then 
                Print( "< objects: ", c!.objects, "\n" );
                gp := c!.magma;
                Print( "    group: " );
                if HasName( gp ) then
                    Print( gp, " = <", GeneratorsOfGroup( gp ), "> >\n" );
                else
                    Print( gp, " >\n" );
                fi;
            else
                Print( "<     objects: ", c!.objects, "\n" ); 
                Print( "   parent gpd: ", Parent( c ), "\n" ); 
                rgp := c!.magma; 
                Print( "   root group: " );
                if HasName( rgp ) then
                    Print( rgp, " = <", GeneratorsOfGroup( rgp ), ">\n" );
                else
                    Print( rgp, "\n" );
                fi; 
                Print( "  conjugators: ", c!.rayConjugators, "\n" );
            fi; 
        od;
    fi;
end );

#############################################################################
##
#M  \=( <g1>, <g2> )  . . . . . . . . . . . . test if two groupoids are equal
##
InstallMethod( \=, "for a connected groupoid", true, ## IsIdenticalObj,
    [ IsGroupoid and IsSinglePiece, IsGroupoid ], 
function ( g1, g2 )

    ## Print( "### method 1 for =\n" ); 
    if not IsSinglePiece( g2 ) then
        return false;
    fi; 
    if not ( IsDirectProductWithCompleteGraph( g1 ) = 
             IsDirectProductWithCompleteGraph( g2 ) ) then
        return false;
    fi;
    if IsDirectProductWithCompleteGraph( g1 ) then
        return ( ( g1!.objects = g2!.objects ) and ( g1!.magma = g2!.magma ) );
    else return ( ( Parent( g1 ) = Parent( g2 ) ) and 
                  ( g1!.magma = g2!.magma ) and 
                  ( g1!.rayConjugators = g2!.rayConjugators ) );
    fi;
end );

InstallMethod( \=, "for a groupoid", true, [ IsGroupoid, IsGroupoid ], 
function ( g1, g2 )
    local  c1, c2, len, obj, i, j;

    ## Print( "### method 2 for =\n" ); 
    c1 := Pieces( g1 );
    c2 := Pieces( g2 );
    len := Length( c1 );
    if ( ( len <> Length(c2) ) or ( ObjectList(g1) <> ObjectList(g2) ) ) then
        return false;
    fi;
    for i in [1..len] do
        obj := c1[i]!.objects[1];
        j := PieceNrOfObject( g2, obj );
        if ( c1[i] <> c2[j] ) then
            return false;
        fi;
    od;
    return true;
end );

#############################################################################
##
#M  Size 
##
InstallOtherMethod( Size, "generic method for a groupoid", true,
    [ IsGroupoid ], 0,
function( gpd )

    local  c, s, sg;
    s := 0;
    for c in Pieces( gpd ) do
        if ( s <> infinity ) then 
            if IsDirectProductWithCompleteGraph( c ) then 
                sg := Size( c!.magma );
            else 
                sg := Size( c!.magma );
            fi;
            if ( sg = infinity ) then
                s := infinity;
            else
                s := s + Length( c!.objects )^2 * sg;
            fi;
        fi;
    od;
    return s;
end );

############################################################################# 
## 
#M  ObjectGroup
## 
InstallMethod( ObjectGroup, "generic method for groupoid and object",
    true, [ IsGroupoid and IsSinglePiece, IsScalar ], 0,
function( G, obj )

    local  pos, id, c, rgp; 

    if not ( obj in G!.objects ) then
        Error( "obj not an object of G" );
    fi;
    if IsDirectProductWithCompleteGraph( G ) then 
        return G!.magma;
    fi; 
    if HasObjectGroups( G ) then 
        pos := Position( G!.objects, obj ); 
        return ObjectGroups(G)[ pos ];
    else 
        rgp := G!.magma; 
        pos := Position( G!.objects, obj );
        c := G!.rayConjugators[ pos ];
        if ( c = One( rgp ) ) then
            return rgp;
        else
            return rgp^c;
        fi;
    fi;
end );

InstallMethod( ObjectGroup, "generic method for groupoid and object",
    true, [ IsGroupoid, IsScalar ], 0,
function( G, obj )

    local  nC, C;

    if not ( obj in ObjectList( G ) ) then
        Error( "obj not an object of G" );
    fi;
    nC := PieceNrOfObject( G, obj ); 
    C := Pieces( G )[ nC ];
    return ObjectGroup( C, obj );
end );

############################################################################# 
## 
#M  ObjectGroups
## 
InstallMethod( ObjectGroups, "generic method for groupoid", true,
    [ IsGroupoid and IsSinglePiece ], 0,
function( gpd )
    return List( gpd!.objects, o -> ObjectGroup( gpd, o ) );
end );

InstallMethod( ObjectGroups, "generic method for groupoid", true,
    [ IsGroupoid ], 0,
function( gpd )
    return List( Pieces( gpd ), C -> ObjectGroups( C ) );
end );



## ======================================================================= ##
##                           Groupoid Elements                             ##
## ======================================================================= ##

#############################################################################
##
#M  IsGroupoidElement
##
InstallMethod( IsGroupoidElement, "generic method for mwo",
    true, [ IsMultiplicativeElement ], 0,
function( e ); 
    return IsPerm( e![1] ); 
    #? (03/10/08)  should also allow for other sorts of groups??
end );

#############################################################################
##
#M  GroupoidElementNC 
#M  GroupoidElement 
##
InstallMethod( GroupoidElementNC, "generic method for a groupoid element",
    true, [ IsMultiplicativeElement, IsScalar, IsScalar ], 0,
function( g, i, j )

    local  fam, filter, e;
     
    fam := FamilyObj( [ g, i, j ] );
    filter := IsMultiplicativeElementWithObjects;
    e := Objectify( NewType( fam, filter ), [ g, i, j ] );
    SetIsGroupoidElement( e, true );
    return e;
end );

InstallMethod( GroupoidElement, "generic method for a groupoid element",
    true, [ IsGroupoid, IsMultiplicativeElement, IsScalar, IsScalar ], 0,
function( gpd, g, i, j )

    local  comp, obs, ok, rays;

    if ( HasIsSinglePiece( gpd ) 
         and IsSinglePiece( gpd ) ) then 
        comp := gpd; 
    else 
        comp := PieceOfObject( gpd, i );
    fi; 
    obs := comp!.objects; 
    if ( not ( i in obs ) or not ( j in obs ) ) then 
        Info( InfoGpd, 1, "one of objects i,j not in the groupoid?" );
    fi;  
    if ( HasIsDirectProductWithCompleteGraph( comp ) 
         and IsDirectProductWithCompleteGraph( comp ) ) then 
        ok := ( g in comp!.magma ); 
    else 
        rays := comp!.rayConjugators; 
        ok := rays[Position(obs,i)] * g * rays[Position(obs,j)]^(-1) 
                in comp!.magma; 
    fi; 
    if not ok then 
        return fail;
    else
        return GroupoidElementNC( g, i, j );
    fi;
end );

#############################################################################
##
#M  PrintObj
#M  ViewObj
##
##  InstallMethod( PrintObj, "for a groupoid element",
##      [ IsMultiplicativeElementWithObjects ],
##  function ( e )
##      Print( "[", e![1], " : ", e![2], " -> ", e![3], "]" );
##  end );
##  
##  InstallMethod( ViewObj, "for a groupoid element",
##      [ IsMultiplicativeElementWithObjects ],
##  function ( e )
##      Print( "[", e![1], " : ", e![2], " -> ", e![3], "]" );
##  end );

#############################################################################
##
#M  \*( <e1>, <e2> )  . . . . . . . . . . .  product of two groupoid elements
#M  \^( <e>, <n> )  . . . . . . . . . . . . . . . power of a groupoid element
#M  \^( <e1>, <e2> )  . . . . .  conjugate of one groupoid element by another
##
InstallMethod( \*, "for two groupoid elements", IsIdenticalObj,
    [ IsGroupoidElement, IsGroupoidElement ], 0,
function( e1, e2 )
    if ( e1![3] <> e2![2] ) then
        return fail;
    fi;
    return GroupoidElementNC( e1![1]*e2![1], e1![2], e2![3] );
end );

InstallMethod( \^, "for a groupoid element and an integer", true,
    [ IsGroupoidElement, IsInt ], 0,
function( e, n )
    local  t, h;

    t := e![2];
    h := e![3];
    if ( n = 1 ) then
        return e;
    elif ( n = -1 ) then
        return GroupoidElementNC( e![1]^-1, h, t );
    elif ( t = h ) then
        return GroupoidElementNC( e![1]^n, t, h );
    fi; 
    return fail; 
end );

##  This operator now (05/03/08) follows the definitions in preprint 07.10

##  InstallOtherMethod( \^, "for two groupoid elements", IsIdenticalObj, 
InstallMethod( ConjugateGroupoidElement, "for two groupoid elements", 
    IsIdenticalObj, 
    [IsMultiplicativeElementWithObjects,IsMultiplicativeElementWithObjects], 0,
function( e1, e2 )
    local  c, p, q;
    ## gpd := FamilyObj( e1 )!.mwo; 
    c := e2![1]; 
    p := e2![2]; 
    if ( e2![3] = p ) then  
        if ( e1![2] = p ) then 
#? (18/09/09) : replaced lots of ....WithObjectsNC( gpd, .., .., .. ); 
#?                          with GroupoidElementNC
            if ( e1![3] = p ) then 
                return GroupoidElementNC( e1![1]^c, p, p ); 
            else 
                return GroupoidElementNC( c^(-1)*e1![1], p, e1![3] ); 
            fi; 
        elif ( e1![3] = p ) then 
            return GroupoidElementNC( c*e2![1], e1![2], p ); 
        else 
            return e1; 
        fi; 
    else 
        q := e2![3]; 
        if ( e1![2] = p ) then 
            if ( e1![3] = p ) then
                return GroupoidElementNC( e1![1]^c, q, q ); 
            elif ( e1![3] = q ) then 
                return GroupoidElementNC( c^(-1)*e1![1]*c^(-1), q, p ); 
            else 
                return GroupoidElementNC( c^(-1)*e1![1], q, e1![3] ); 
            fi; 
        elif ( e1![2] = q ) then 
            if ( e1![3] = q ) then 
                return GroupoidElementNC( e1![1]^(c^(-1)), p, p ); 
            elif ( e1![3] = p ) then 
                return GroupoidElementNC( c*e1![1]*c, p, q ); 
            else 
                return GroupoidElementNC( c*e1![1], p, e1![3] ); 
            fi; 
        elif ( e1![3] = p ) then 
            return GroupoidElementNC( e1![1]*c, e1![2], q ); 
        elif ( e1![3] = q ) then 
            return GroupoidElementNC( e1![1]*c^(-1), e1![2], p ); 
        else 
            return e1; 
        fi;
    fi; 
end );

#############################################################################
##
#M  \^( <gpd>, <elt> )
##
##  InstallOtherMethod( \^, "generic method for groupoid and element",
##      IsCollsElms, [ IsGroupoid, IsGroupoidElement ], ConjugateGroupoid );

#############################################################################
##
##  added 05/03/08 : need version for groupoids with several pieces 
## 
#M  ConjugateGroupoid( <gpd>, <elt> )
#M  ConjugateSubgroupoid( <gpd>, <elt> )
##
##  InstallMethod( ConjugateGroupoid, "<gpd>, <elt>", true,
##      [ IsGroupoid and IsSinglePiece, IsGroupoidElement ],
##      function( gpd, elt )
##  
##      local  gens, ims, conj;
##  
##      gens := GeneratorsOfGroupoid( gpd ); 
##      ims := List( gens, g -> g^elt );
##      conj := SinglePieceGroupoidByGenerators( ims );
    ## leftovers from the version for groups: 
    ## OnTuples( GeneratorsOfGroupoid( gpd ), elt ), One( gpd ) );
    ## UseIsomorphismRelation( gpd, conj );
##      return conj;
##  end );

##  InstallMethod( ConjugateSubgroupoid, 
##      "for groupoid with parent, and group element", true, 
##      [IsGroupoid and HasParent, IsGroupoidElement],
##  function( gpd, elt ) 
##      local  conj; 
##      conj := ConjugateGroupoid( gpd, elt );
##      if not IsIdenticalObj( Parent(gpd), conj ) then
##          SetParent( conj, Parent( gpd ) );
##      fi;
##      return conj;
##  end );

##  InstallOtherMethod( ConjugateSubgroupoid, "for groupoid without parent",
##    true, [ IsGroupoid, IsMultiplicativeElementWithObjects], ConjugateGroupoid );

#############################################################################
##
#M  <e> in <G> 
##
#?  (12/09/08)  tried again - but no joy! 

InstallOtherMethod( \in, "for groupoid element and a standard groupoid", 
    IsElmsColls, 
    [ IsGroupoidElement, IsGroupoid and IsDirectProductWithCompleteGraph ], 0,
function( e, gpd )
    local  obs;
    obs := gpd!.objects; 
    return ( (e![1] in gpd!.magma) and (e![2] in obs) and (e![3] in obs) );
end );

InstallMethod( IsElementInGroupoid, 
    "for groupoid element and a standard groupoid", true, 
    [ IsGroupoidElement, IsGroupoid and IsSinglePiece ], 0,
function( e, gpd )
    local  obs, r1, r2, rays;
    obs := gpd!.objects; 
    if not ( (e![2] in obs) and (e![3] in obs) ) then 
        return false; 
    fi; 
    if ( HasIsDirectProductWithCompleteGraph( gpd ) 
         and IsDirectProductWithCompleteGraph( gpd ) ) then 
        return (e![1] in gpd!.magma);
    else 
        rays := gpd!.rayConjugators; 
        r1 := rays[ Position( obs, e![2] ) ]; 
        r2 := rays[ Position( obs, e![3] ) ]; 
        return ( r1 * e![1] * r2^(-1) in gpd!.magma ); 
    fi; 
end );

InstallMethod( IsElementInGroupoid, 
    "for groupoid element and a union of constituents", true, 
    [ IsGroupoidElement, IsGroupoid and HasPieces ], 0,
function( e, gpd )
    return IsElementInGroupoid( e, PieceOfObject( gpd, e![2] ) ); 
end );

##  InstallMethod( \in, "for a groupoid element and a groupoid",
##      true, [ IsGroupoidElement, IsGroupoid ], 0,
##  function( e, gpd )
##      local  comp;
##      comp := PieceOfObject( gpd, e![2] );
##  Print("comp = ",comp,"\n");
##      if ( comp = fail ) then
##          return false;
##      else
##          return ( e in comp );
##      fi;
##  end );

############################################################################# 
## 
#M  IdentityElement
## 
InstallMethod( IdentityElement, "for a connected groupoid and object",
    true, [ IsGroupoid and IsSinglePiece, IsScalar ], 0,
function( gpd, obj )
    return GroupoidElementNC( One(gpd!.magma), obj, obj );
end );

InstallMethod( IdentityElement, "generic method for groupoid and object",
    true, [ IsGroupoid, IsScalar ], 0,
function( gpd, obj )
    local  comp;
    comp := PieceOfObject( gpd, obj );
    return GroupoidElementNC( One(comp!.magma), obj, obj );
end );

#############################################################################
##
#M  PrintObj 
##
InstallMethod( PrintObj, "for a subset of elements of a groupoid", true, 
    [ IsHomsetCosets ], 0,
function ( hc )
    
    local  iter, g;
    if ( hc![6] = "s" ) then 
        Print( "<star at ", hc![3], " with group ", hc![2], ">" );
    elif ( hc![6] = "c" ) then 
        Print( "<costar at ", hc![4], " with group ", hc![2], ">" );
    elif ( hc![6] = "h" ) then 
        Print( "<homset ", hc![3][1], " -> ", hc![4][1],
               " with group ", hc![2], ">" );
    elif ( hc![6] in [ "r", "l", "d" ] ) then 
        Print( "homset-cosets for coset" );
    else
        Print( "<object>");
    fi; 
end );

#############################################################################
##
#M  Iterator( <cset> ) . . . . . . . . . . . . . iterator for groupoid coset 
##
InstallMethod( Iterator, "for a groupoid coset", [ IsGroupoidCoset ], 
    cset -> Iterator( HomsetCosetsGroupoidCoset( cset ) ) ); 

#############################################################################
##
#M  \=( <cset> ) . . . . . . . . . . . . . . . . . . . . for groupoid cosets 
##
InstallMethod( \=, "for groupoid cosets", [IsGroupoidCoset, IsGroupoidCoset], 
function( c1, c2 ) 
    return ( ( Representative(c1) = Representative(c2) ) 
           and ( ActingDomain(c1) = ActingDomain(c2) ) 
        and ( SuperDomain(c1) = SuperDomain(c2) ) 
     and (HomsetCosetsGroupoidCoset(c1)![6]=HomsetCosetsGroupoidCoset(c2)![6])
    );
end );

#############################################################################
##
#M  Size( <homsets> ) . . . . . . . . size for star, costar, homset or coset
##
InstallMethod( Size, "for a subset of a connected groupoid", 
    [ IsHomsetCosets ], 
function( hc )
    return Size( hc![2]) * Length(hc![3]) * Length(hc![4]); 
end );

#############################################################################
##
#M  Iterator( <homsets> ) . . . . iterator for star, costar, homset or coset
##
InstallMethod( Iterator, "for a subset of a connected groupoid", 
    [ IsHomsetCosets ], 
function( hc )
    return IteratorByFunctions( rec( 
        IsDoneIterator := function( iter )
            return ( IsDoneIterator( iter!.groupIterator ) 
                     and ( iter!.tpos = iter!.tlen )
                     and ( iter!.hpos = iter!.hlen ) );
            end, 
        NextIterator := function( iter )
            if ( iter!.tpos = 0 ) then
                iter!.gpelt := NextIterator( iter!.groupIterator );
                iter!.tpos := 1;
                iter!.hpos := 1;
           elif ((iter!.tpos = iter!.tlen) and (iter!.hpos = iter!.hlen)) then 
                iter!.gpelt := NextIterator( iter!.groupIterator );
                iter!.tpos := 1;
                iter!.hpos := 1;
            elif ( iter!.hpos = iter!.hlen ) then
                iter!.hpos := 1;
                iter!.tpos := iter!.tpos + 1;
            else 
                iter!.hpos := iter!.hpos + 1;
            fi; 
            if ( hc![6] in [ "c", "r" ] ) then 
                return GroupoidElementNC( 
                    #? iter!.fgpd!.mwo, 
                        (hc![5][iter!.tpos])*(iter!.gpelt), 
                        iter!.tobs[iter!.tpos], iter!.hobs[iter!.hpos] );
            elif ( hc![6] in [ "s", "h" ] ) then 
                return GroupoidElementNC( 
                    #? iter!.fgpd!.mwo, 
                        (iter!.gpelt)*(hc![5][iter!.hpos]), 
                        iter!.tobs[iter!.tpos], iter!.hobs[iter!.hpos] );
            elif ( hc![6] in [ "l" ] ) then 
                return GroupoidElementNC( 
                    #? iter!.fgpd!.mwo, 
                        (iter!.gpelt)^(-1)*(hc![5][iter!.hpos]), 
                        iter!.tobs[iter!.tpos], iter!.hobs[iter!.hpos] );
            elif ( hc![6] in [ "d" ] ) then 
                Error( "double cosets not yet implemented" ); 
            fi; 
            end,
        ShallowCopy := iter -> 
            rec( groupIterator := ShallowCopy( iter!.groupIterator ), 
                 fgpd := iter!.fgpd, 
                 gpelt := iter!.gpelt,
                 tobs := iter!.tobs,
                 tlen := iter!.tlen,
                 hobs := iter!.hobs,
                 hlen := iter!.hlen,
                 tpos := iter!.tpos,
                 hpos := iter!.hpos ),
        groupIterator := Iterator( hc![2] ), 
        fgpd := hc![1], 
        gpelt := 0,
        tobs := hc![3],
        tlen := Length( hc![3] ),
        hobs := hc![4],
        hlen := Length( hc![4] ),
        tpos := 0,
        hpos := 0 ) );
end );

#############################################################################
##
#M  ObjectStarNC 
#M  ObjectStar 
##
InstallMethod( ObjectStarNC, "for a connected groupoid and an object",
    true, [ IsGroupoid and IsSinglePiece, IsScalar ], 0,
function( gpd, obj )

    local  gp, obs, nobs, fam, filter, st, fgpd, rays, pos, rpos;

    obs := gpd!.objects; 
    filter := IsHomsetCosetsRep; 
    if ( HasIsDirectProductWithCompleteGraph( gpd ) 
         and IsDirectProductWithCompleteGraph( gpd ) ) then 
        gp := gpd!.magma; 
        fgpd := FamilyObj( gpd ); 
        rays := List( obs, o -> One(gp) ); 
    else 
        fgpd := FamilyObj( Parent( gpd ) ); 
        gp := ObjectGroup( gpd, obj ); 
        rays := gpd!.rayConjugators; 
        pos := Position( obs, obj ); 
        if ( pos <> 1 ) then  ## not the root object 
            rpos := rays[pos]^(-1); 
            rays := List( [1..Length(obs)], j -> rpos*rays[j] ); 
        fi; 
    fi; 
    fam := FamilyObj( [ fgpd, gp, [obj], obs, rays, "s" ] ); 
    st := Objectify( NewType( fam, filter ), 
                     [ fgpd, gp, [obj], obs, rays, "s" ]);
    SetIsHomsetCosets( st, true ); 
    return st;
end );

InstallMethod( ObjectStar, "generic method for a groupoid and an object",
    true, [ IsGroupoid, IsScalar ], 0,
function( gpd, obj )

    local  comp;
    if ( IsSinglePiece(gpd) and ( obj in ObjectList(gpd) ) ) then
        return ObjectStarNC( gpd, obj );
    else
        comp := PieceOfObject( gpd, obj );
        if ( comp = fail ) then
            Info( InfoGpd, 1, "obj not an object in gpd" );
            return fail;
        else
            return ObjectStarNC( comp, obj );
        fi;
    fi;
end );

#############################################################################
##
#M  ObjectCostarNC 
#M  ObjectCostar 
##
InstallMethod( ObjectCostarNC, "for a connected groupoid and an object",
    true, [ IsGroupoid and IsSinglePiece, IsScalar ], 0,
function( gpd, obj )

    local  gp, obs, nobs, fam, filter, cst, fgpd, rays, pos, rpos;

    obs := gpd!.objects; 
    filter := IsHomsetCosetsRep; 
    if ( HasIsDirectProductWithCompleteGraph( gpd ) 
         and IsDirectProductWithCompleteGraph( gpd ) ) then 
        gp := gpd!.magma; 
        fgpd := FamilyObj( gpd ); 
        rays := List( obs, o -> One(gp) ); 
    else 
        fgpd := FamilyObj( Parent( gpd ) ); 
        gp := ObjectGroup( gpd, obj ); 
        rays := List( gpd!.rayConjugators, r -> r^(-1) ); 
        pos := Position( obs, obj ); 
        if ( pos <> 1 ) then  ## not the root object 
            rpos := rays[pos]^(-1); 
            rays := List( [1..Length(obs)], j -> rays[j]*rpos ); 
        fi; 
    fi; 
    fam := FamilyObj( [ fgpd, gp, obs, [obj], rays, "c" ] ); 
    cst := Objectify( NewType( fam, filter ), 
                      [ fgpd, gp, obs, [obj], rays, "c" ] );
    SetIsHomsetCosets( cst, true ); 
    return cst;
end );

InstallMethod( ObjectCostar, "generic method for a groupoid and an object",
    true, [ IsGroupoid, IsScalar ], 0,
function( gpd, obj )

    local  comp;
    if ( IsSinglePiece(gpd) and ( obj in ObjectList(gpd) ) ) then
        return ObjectCostarNC( gpd, obj );
    else
        comp := PieceOfObject( gpd, obj );
        if ( comp = fail ) then
            Info( InfoGpd, 1, "obj not an object in gpd" );
            return fail;
        else
            return ObjectCostarNC( comp, obj );
        fi;
    fi;
end );

#############################################################################
##
#M  HomsetNC 
#M  Homset
##
InstallMethod( HomsetNC, "for a connected groupoid and two objects",
    true, [ IsGroupoid and IsSinglePiece, IsScalar, IsScalar ], 0,
function( gpd, o1, o2 )

    local  gp, obs, fgpd, fam, filter, rays, gen, hs, p1, p2;
    
    obs := gpd!.objects; 
    filter := IsHomsetCosetsRep; 
    if ( HasIsDirectProductWithCompleteGraph( gpd ) 
         and IsDirectProductWithCompleteGraph( gpd ) ) then 
        gp := gpd!.magma;
        fgpd := FamilyObj( gpd ); 
        gen := [ One(gp) ]; 
    else 
        fgpd := FamilyObj( Parent( gpd ) ); 
        gp := ObjectGroup( gpd, o1 ); 
        p1 := Position( obs, o1 ); 
        p2 := Position( obs, o2 );
        gen := [ gpd!.rayConjugators[p1]^(-1) * gpd!.rayConjugators[p2] ]; 
    fi; 
    fam := FamilyObj( [ fgpd, gp, [o1], [o2], gen, "h" ] ); 
    hs := Objectify( NewType( fam, filter ), 
                     [ fgpd, gp, [o1], [o2], gen, "h" ] );
    SetIsHomsetCosets( hs, true ); 
    return hs;
end );

InstallMethod( Homset, "generic method for a groupoid and two objects",
    true, [ IsGroupoid, IsScalar, IsScalar ], 0,
function( gpd, o1, o2 )

    local  obs, comp;

    obs := ObjectList( gpd );
    if not ( ( o1 in obs ) and ( o2 in obs ) ) then
        Info( InfoGpd, 1, "o1,o2 not objects in gpd" );
        return fail;
    fi;
    if IsSinglePiece( gpd ) then
        return HomsetNC( gpd, o1, o2 );
    else
        comp := PieceOfObject( gpd, o1 );
        if not ( o2 in comp!.objects ) then
            Info( InfoGpd, 1, "o1,o2 not objects in same constituent" );
            return fail;
        else
            return HomsetNC( comp, o1, o2 );
        fi;
    fi;
end );

#############################################################################
##
#M  ElementsOfGroupoid
##
InstallMethod( ElementsOfGroupoid, "for a connected groupoid", true,
    [ IsGroupoid and IsSinglePiece ], 0,
function( gpd )

    local  iter, elts;

    elts := [ ];
    iter := Iterator( gpd );
    while not IsDoneIterator( iter ) do
        Add( elts, NextIterator( iter ) );
    od;
    return elts;
end );

InstallMethod( ElementsOfGroupoid, "generic method for a groupoid", true,
    [ IsGroupoid ], 0,
function( gpd )

    local  comps, c, elts;
    comps := Pieces( gpd );
    elts := [ ];
    for c in comps do
        Append( elts, ElementsOfGroupoid( c ) );
    od;
    return elts;
end );

#############################################################################
##
#M  Iterator( <gpd> ) . . . . . . . . . . . . . . . . iterator for a groupoid
##
InstallMethod( Iterator, "for a connected groupoid", 
    [ IsGroupoid and IsSinglePiece ], 
function( gpd )
    return IteratorByFunctions( rec( 
        IsDoneIterator := function( iter )
            return ( IsDoneIterator( iter!.groupIterator ) 
                     and ( iter!.tpos = iter!.len )
                     and ( iter!.hpos = iter!.len ) );
            end, 
        NextIterator := function( iter )
            if ( iter!.tpos = 0 ) then
                iter!.gpelt := NextIterator( iter!.groupIterator );
                iter!.tpos := 1;
                iter!.hpos := 1;
            elif ((iter!.tpos = iter!.len) and (iter!.hpos = iter!.len)) then 
                iter!.gpelt := NextIterator( iter!.groupIterator );
                iter!.tpos := 1;
                iter!.hpos := 1;
            elif ( iter!.hpos = iter!.len ) then
                iter!.hpos := 1;
                iter!.tpos := iter!.tpos + 1;
            else 
                iter!.hpos := iter!.hpos + 1;
            fi;
            if ( HasIsDirectProductWithCompleteGraph( gpd ) 
                 and IsDirectProductWithCompleteGraph( gpd ) ) then 
               return GroupoidElementNC( 
                  #? FamilyObj(gpd)!.mwo, 
                  iter!.gpelt, iter!.obs[iter!.tpos], iter!.obs[iter!.hpos] );
            else 
               return GroupoidElementNC( 
                  #? FamilyObj( Parent(gpd) )!.mwo, 
                    gpd!.rayConjugators[iter!.tpos]^(-1) 
                      * iter!.gpelt * gpd!.rayConjugators[iter!.hpos], 
                      iter!.obs[iter!.tpos], iter!.obs[iter!.hpos] );
            fi; 
            end, 
        ShallowCopy := iter -> 
            rec( groupIterator := ShallowCopy( iter!.groupIterator ), 
                 gpelt := iter!.gpelt,
                 obs := iter!.obs,
                 len := iter!.len,
                 tpos := iter!.tpos,
                 hpos := iter!.hpos ),
        groupIterator := Iterator( gpd!.magma ), 
        gpelt := 0, 
        obs := gpd!.objects,
        len := Length( gpd!.objects ),
        tpos := 0,
        hpos := 0 ) );
end );

InstallMethod( Iterator, "generic method for a groupoid", [ IsGroupoid ], 
function( gpd )
    return IteratorByFunctions( rec( 
        IsDoneIterator := function( iter )
            return ( IsDoneIterator( iter!.groupoidIterator ) 
                     and ( iter!.cpos = iter!.len ) );
            end, 
        NextIterator := function( iter )
            if IsDoneIterator( iter!.groupoidIterator ) then 
                iter!.cpos := iter!.cpos + 1;
                iter!.groupoidIterator := 
                    Iterator( iter!.constituents[iter!.cpos] );
            fi;
            return NextIterator( iter!.groupoidIterator );
            end,
        ShallowCopy := iter -> 
            rec( constituents := iter!.constituents,
                 len := iter!.len,
                 groupoidIterator := ShallowCopy( iter!.groupoidIterator ),
                 cpos := iter!.cpos ),
        constituents := Pieces( gpd ),
        len := Length( Pieces( gpd ) ),
        groupoidIterator := Iterator( Pieces( gpd )[1] ),
        cpos := 1 ) );
end );

## ======================================================================= ##
##                                Subgroupoids                             ##
## ======================================================================= ##

#############################################################################
##
#F  Subgroupoid( <gpd>, <subgp> )        connected groupoids
#F  Subgroupoid( <gpd>, <gps>, <obs> )   discrete subgroupoid 
#F  Subgroupoid( <gpd>, <comp> )         subgroupoid as list of constituents
##
InstallGlobalFunction( Subgroupoid, function( arg )

    local  nargs, gpd, id, rays, gp, sub;
    nargs := Length( arg );
    gpd := arg[1];
    if not IsGroupoid( gpd ) then 
        Info( InfoGpd, 1, "arg[1] is not a groupoid" );
        return fail;
    fi;
    # by subgroup
    if ( ( nargs = 2 ) and IsSinglePiece( arg[1] )
                       and IsGroup( arg[2] ) ) then
        gp := gpd!.magma;
        Info( InfoGpd, 2, "connected subgroupoid" );
        sub := SinglePieceGroupoid( arg[2], gpd!.objects );
        SetParentAttr( sub, gpd );
        return sub;
    fi;
    # discrete subgroupoid
    if ( ( nargs = 3 ) and IsHomogeneousList( arg[2] ) 
                       and IsHomogeneousList( arg[3] ) ) then
        Info( InfoGpd, 2, "discrete subgroupoid" );
        return DiscreteSubgroupoid( arg[1], arg[2], arg[3] );
    fi;
    # list of constituents
    if ( ( nargs = 2 ) and IsHomogeneousList( arg[2] ) ) then
        Info( InfoGpd, 3, "subgroupoid by constituents" );
        return SubgroupoidByPieces( arg[1], arg[2] );
    fi;
    Info( InfoGpd, 1, SUB_CONSTRUCTORS );
    return fail;
end );

#############################################################################
##
#F  IsSubgroupoid( <G>, <U> )
##
InstallMethod( IsSubgroupoid, "generic method for two groupoids", true,
    [ IsGroupoid, IsGroupoid], 0,
function( G, U ) 
    ## Print( "IsSubgroupoid 1 : ", G, " >= ", U, "\n" ); 
    if ( HasParentAttr( U ) and ( ParentAttr( U ) = G ) ) then 
        return true;
    fi;
    if not IsSubset( ObjectList(G), ObjectList(U) ) then
        return false;
    fi; 
    return ForAll( Pieces( U ), C -> IsSubgroupoid( G, C ) );
end );

InstallMethod( IsSubgroupoid, "generic method for two groupoids", true,
    [ IsGroupoid, IsGroupoid and IsSinglePiece], 0,
function( G, U )
    ## Print( "IsSubgroupoid 2 : ", G, " >= ", U, "\n" ); 
    if ( HasParentAttr( U ) and ( ParentAttr( U ) = G ) ) then 
        return true;
    fi;
    if not IsSubset( ObjectList(G), ObjectList(U) ) then
        return false;
    fi; 
    return IsSubgroupoid( PieceOfObject(G,U!.objects[1]), U ); 
end ); 

InstallMethod( IsSubgroupoid, "generic method for two groupoids", true,
  [IsGroupoid and IsSinglePiece, IsGroupoid and IsSinglePiece], 0,
function( G, U )

    local  objG, objU;
    ## Print( "IsSubgroupoid 3 : ", G, " >= ", U, "\n" ); 
    if ( HasParentAttr( U ) and ( ParentAttr( U ) = G ) ) then 
        return true;
    fi;
    if not IsSubset( ObjectList(G), ObjectList(U) ) then
        return false;
    fi;  
    if ( G = U ) then 
        return true; 
    fi; 
    objG := G!.objects; 
    objU := U!.objects; 
    if not IsSubset( objG, objU ) then
        return false;
    fi; 
    if not IsSubgroup( ObjectGroup( G, objU[1] ), U!.magma ) then
        return false; 
    fi; 
    if not HasParentAttr( U ) then 
        SetParent( U, G ); 
    fi; 
    return true; 
end ); 

#############################################################################
##
#M  SubgroupoidBySubgroup
##
InstallMethod( SubgroupoidBySubgroup,
    "generic method for direct prod with complete graph and subgroup", true,
    [ IsGroupoid and IsDirectProductWithCompleteGraph, IsGroup ], 0,
function( gpd, sgp )
    local  sub, gp; 
    gp := gpd!.magma; 
    if not IsSubgroup( gp, sgp ) then 
        Error( "sgp is not a subgroup of gpd!.magma" ); 
    fi; 
    sub := SinglePieceGroupoid( sgp, gpd!.objects );
    SetParentAttr( sub, gpd );
    return sub; 
end ); 

#############################################################################
##
#M  SubgroupoidByPieces
##
InstallMethod( SubgroupoidByPieces,
    "generic method for a groupoid and a list of pieces", true,
    [ IsGroupoid, IsHomogeneousList ], 0, 
function( gpd, pieces )

    local  pieceU, U, c, piece;

    pieceU := List( pieces, c -> SinglePieceGroupoid( c[1], c[2] ) ); 
    if ( Length( pieceU ) > 1 ) then
        U := UnionOfPieces( pieceU );
    else
        U := pieceU[1];
    fi;
    if not IsSubgroupoid( gpd, U ) then 
        Info( InfoGpd, 1, "union of pieces is not a subgroupoid of gpd" );
        return fail;
    fi;
    if ForAll( pieces, c -> ( Length( c[2] ) = 1 ) ) then
        SetIsDiscreteDomain( U, true );
    else
        SetIsDiscreteDomain( U, false );
    fi;
    SetParentAttr( U, gpd );
    for c in pieceU do 
        piece := PieceOfObject( gpd, c!.objects[1] ); 
        SetParentAttr( c, FullSubgroupoid( piece, ObjectList(c) ) );
    od;
    return U;
end );

#############################################################################
##
#M  DiscreteSubgroupoid
##
InstallMethod( DiscreteSubgroupoid, "generic method for a groupoid", true,
    [ IsGroupoid, IsHomogeneousList, IsHomogeneousList ], 0,
function( G, obs, gps )

    local  pieceU, U, len, o, pieceG, obsg, C, i;

    obsg := ObjectList( G );
    len := Length( obs );
    pieceU := ListWithIdenticalEntries( len, 0 );
    if not ( len = Length( gps ) ) then
        Error( "object and subgroup lists not of same length" );
    fi;
    for i in [1..len] do
        o := obs[i];
        if not ( o in obsg ) then
            Error( "subgroupoid object not in groupoid" );
        fi;
        C := PieceOfObject( G, o );
        if not IsSubgroup( C!.magma, gps[i] ) then
            Error( "not a subgroup of object group" );
        fi;
        pieceU[i] := [ [o], gps[i] ];
    od;
    U := SubgroupoidByPieces( G, pieceU );
    SetIsDiscreteDomain( U, true );
    SetParentAttr( U, G );
    return U;
end );

#############################################################################
##
#M  MaximalDiscreteSubgroupoid
##
InstallMethod( MaximalDiscreteSubgroupoid,
    "generic method for a groupoid", true, [ IsGroupoid ], 0,
function( gpd )

    local  obs, len, gps, i, piece;

    obs := ObjectList( gpd );
    len := Length( obs );
    gps := ListWithIdenticalEntries( len, 0 );
    for i in [1..len] do
        piece := PieceOfObject( gpd, obs[i] );
        gps[i] := piece!.magma;
    od;
    return DiscreteSubgroupoid( gpd, obs, gps );
end );

#############################################################################
##
#M  FullSubgroupoid
#M  FullIdentitySubgroupoid
#M  DiscreteIdentitySubgroupoid
##
InstallMethod( FullSubgroupoid, "for a connected groupoid and set of objects", 
    true, [ IsGroupoid and IsSinglePiece, IsHomogeneousList ], 0,
function( G, sobs )
    local  sgpd, rgp, rayG, pG, spG, pos, rays, r1; 
    if ( HasIsDirectProductWithCompleteGraph( G ) 
         and IsDirectProductWithCompleteGraph( G ) ) then 
        if not HasIsSSortedList( sobs ) then 
            Sort( sobs ); 
        fi; 
        sgpd :=  SinglePieceGroupoidNC( G!.magma, sobs ); 
        SetParentAttr( sgpd, G ); 
    else 
        pG := Parent( G ); 
        spG := FullSubgroupoid( pG, sobs ); 
        rgp := ObjectGroup( G, sobs[1] );  
        rayG := G!.rayConjugators; 
        pos := List( sobs, j -> Position( G!.objects, j ) ); 
        r1 := rayG[ pos[1] ]^(-1); 
        rays := List( pos, j -> r1*rayG[j] ); 
        sgpd := SinglePieceGroupoidWithRays( spG, rgp, rays );
        SetParentAttr( sgpd, spG ); 
    fi; 
    return sgpd; 
end );

InstallMethod( FullSubgroupoid, "for a groupoid and set of objects", true,
    [ IsGroupoid, IsHomogeneousList ], 0,
function( G, sobs )
    local  c1, c2, len, j, c, obc, sobc;
    c1 := Pieces( G );
    len := Length( c1 );
    c2 := [ ];
    for j in [1..len] do
        c := c1[j];
        obc := c!.objects;
        sobc := Intersection( sobs, obc );
        if ( sobc <> [ ] ) then
            Add( c2, [ sobc, c!.magma ] );
        fi;
    od;
    return SubgroupoidByPieces( G, c2 );
end );

InstallMethod( FullIdentitySubgroupoid, "for a connected groupoid", 
    true, [ IsGroupoid and IsSinglePiece ], 0,
function( gpd ) 
    local  id, sub;
    id := IdentitySubgroup( gpd!.magma );
    sub := SinglePieceGroupoidNC( id, gpd!.objects );
    SetParentAttr( sub, gpd );
    return sub;
end );

InstallMethod( FullIdentitySubgroupoid, "generic method for a groupoid", 
    true, [ IsGroupoid ], 0,
function( gpd )
    local  pieces, subs;
    pieces := Pieces( gpd );
    subs := List( pieces, c -> FullIdentitySubgroupoid(c) );
    return UnionOfPieces( subs );
end );

InstallMethod( DiscreteIdentitySubgroupoid, "for a connected groupoid", 
    true, [ IsGroupoid and IsSinglePiece ], 0,
function( gpd )
    local  id, obs, gps;
    obs := gpd!.objects;
    id := IdentitySubgroup( gpd!.magma );
    gps := List( obs, o -> id );
    return DiscreteSubgroupoid( gpd, obs, gps );
end );

InstallMethod( DiscreteIdentitySubgroupoid, "generic method for a groupoid", 
    true, [ IsGroupoid ], 0,
function( gpd )
    local  pieces, subs;
    pieces := Pieces( gpd );
    subs := List( pieces, c -> DiscreteIdentitySubgroupoid(c) );
    return UnionOfPieces( subs );
end );

#############################################################################
##
#M  PiecePositions
##
InstallMethod( PiecePositions, "for a groupoid and a subgroupoid", true,
    [ IsGroupoid, IsGroupoid ], 0,
function( G, U )

    local  obsG, obU, lenG, lenU, pos, i, oi, j, found, p;

    if IsSinglePiece( G ) then
        if IsSinglePiece( U ) then
            pos := [ 1 ];
        else
            pos := List( Pieces(U), c -> 1 );
        fi;
    else
        obsG := List( Pieces(G), c -> c!.objects );
        lenG := Length( obsG );
        obU := List( Pieces(U), c -> c!.objects[1] );
        lenU := Length( obU );
        pos := ListWithIdenticalEntries( lenU, 0 );
        for i in [1..lenU] do
            oi := obU[i];
            found := false;
            j := 0;
            while ( not found ) do
                j := j+1;
                p := Position( obsG[j], oi );
                if ( p <> fail ) then
                    pos[i] := j;
                    found := true;
                fi;
            od;
        od;
    fi;
    Info( InfoGpd, 3, "constituent positions = ", pos ); 
    return pos;
end );

#############################################################################
##
#M  IsWideSubgroupoid
##
InstallMethod( IsWideSubgroupoid, "for a groupoid and a subgroupoid", true,
    [ IsGroupoid, IsGroupoid ], 0,
function( G, U ) 
    local  ok1, ok2; 
    ok1 := ObjectList( G ) = ObjectList( U ); 
    ok2 := IsSubgroupoid( G, U ); 
     return ( ok1 and ok2 ); 
end );

## ======================================================================== ##
##                            Groupoid Cosets                               ##
## ======================================================================== ##

##############################################################################
##
#M  RightTransversal 
#M  RightCoset 
#M  LeftCoset
#M  DoubleCoset
##
#? (24/09/08)  decided to include the larger groupoid - a mistake ?? 
#? (24/09/08)  should split this into a non-NC and an NC operation ?? 
## (03/10/08)  now allowing U not to be wide in G (see RightCosets). 

InstallOtherMethod( RightCoset, "for groupoid, subgroupoid and element", 
    true,  ##  IsCollsElms ?? 
    [ IsGroupoid, IsGroupoid, IsMultiplicativeElementWithObjects ],
    0, 
function( gpd, sgpd, e ) 

    local  G, U, fam, r, hc; 

    if IsSinglePiece( gpd ) then 
        G := gpd; 
    else
        Info( InfoGpd, 1, "comment: gpd not a single piece!" ); 
        G := PieceOfObject( gpd, e![2] ); 
    fi; 
    if IsSinglePiece( sgpd ) then 
        U := sgpd; 
    else
        Info( InfoGpd, 1, "comment: sgpd not a single piece!" ); 
        U := PieceOfObject( sgpd, e![2] ); 
    fi; 
    if not IsSubdomainWithObjects( G, U ) then 
        Error( "U not a subgroupoid of G" ); 
    fi; 
    fam := FamilyObj( G ); 
    if not IsBound(fam!.rightGroupoidCosetsDefaultSizeType) then
        fam!.rightGroupoidCosetsDefaultSizeType 
            := NewType( fam, IsRightCosetDefaultRep and HasActingDomain 
               and HasFunctionAction and HasRepresentative and HasSize );
    fi;
    r := rec();
    ObjectifyWithAttributes( r, fam!.rightGroupoidCosetsDefaultSizeType, 
        IsGroupoidCoset, true, 
        SuperDomain, G, 
        ActingDomain, U, 
##        FunctionAction, OnLeftInverse,  #? (18/09/08) incorrect! 
        Size, Size( ObjectCostar( U, e![2] ) ), 
        Representative, e ); 
    hc := HomsetCosetsGroupoidCoset( r ); 
    return r; 
end ); 

InstallMethod( LeftCoset, "for groupoid, subgroupoid and element", 
    true,  ##  IsCollsElms ?? 
    [ IsGroupoid, IsGroupoid, IsMultiplicativeElementWithObjects ],
    0, 
function( gpd, sgpd, e ) 

    local  G, U, fam, l, hc; 
    if IsSinglePiece( gpd ) then 
        G := gpd; 
    else
        Info( InfoGpd, 1, "comment: gpd not a single piece!" ); 
        G := PieceOfObject( gpd, e![2] ); 
    fi; 
    if IsSinglePiece( sgpd ) then 
        U := sgpd; 
    else
        Info( InfoGpd, 1, "comment: sgpd not a single piece!" ); 
        U := PieceOfObject( sgpd, e![2] ); 
    fi; 
    if not IsSubdomainWithObjects( G, U ) then 
        Error( "U not a subgroupoid of G" ); 
    fi; 
    fam := FamilyObj( G );
    if not IsBound(fam!.leftGroupoidCosetsDefaultSizeType) then
        fam!.leftGroupoidCosetsDefaultSizeType 
            := NewType( fam, IsMyLeftCosetDefaultRep and HasActingDomain 
               and HasFunctionAction and HasRepresentative and HasSize );
    fi;
    l := rec();
    ObjectifyWithAttributes( l, fam!.leftGroupoidCosetsDefaultSizeType, 
        IsGroupoidCoset, true, 
        SuperDomain, G, 
        ActingDomain, U, 
##        FunctionAction, OnLeftInverse,  #? (18/09/08) incorrect!
        Size, Size( ObjectStar( U, e![3] ) ),   
        Representative, e ); 
    hc := HomsetCosetsGroupoidCoset( l ); 
    return l;
end ); 

InstallMethod( PrintObj, "LeftCoset", true, 
    [ IsGroupoidCoset and IsMyLeftCosetDefaultRep ], 0,
function( l )
  Print( "LeftCoset(", Representative(l), ",", ActingDomain(l), ")" );
end); 

InstallMethod( ViewObj, "LeftCoset", true, 
    [ IsGroupoidCoset and IsMyLeftCosetDefaultRep ], 0,
function( l )
  Print( "LeftCoset(", Representative(l), ",", ActingDomain(l), ")" );
end); 

InstallMethod( HomsetCosetsGroupoidCoset, "for right coset of groupoid",
    true, [ IsGroupoidCoset ], 0,
function( cset )

    local  U, e, G, obs, ogp, o1, o2, gpcset, fgpd, rays, pos, rpos, 
           fam, filter, hc, isrt, C; 
    e := Representative( cset ); 
    G := SuperDomain( cset ); 
    U := ActingDomain( cset ); 
    if not IsElementInGroupoid( e, G ) then 
        Error( "element e not in groupoid G" ); 
    fi; 
    filter := IsHomsetCosetsRep; 
    isrt := IsRightCosetDefaultRep( cset ); 
    o1 := e![2];
    o2 := e![3]; 
    if isrt then 
        ogp := ObjectGroup( U, o1 ); 
    else 
        ogp := ObjectGroup( U, o2 ); 
    fi; 
    filter := IsHomsetCosetsRep; 
    if ( HasIsSinglePiece(U) and IsSinglePiece(U) ) then 
        C := U; 
    elif isrt then 
        C := PieceOfObject( U, o1 ); 
    else 
        C := PieceOfObject( U, o2 ); 
    fi; 
    obs := C!.objects;
    if ( HasIsDirectProductWithCompleteGraph( C ) 
         and IsDirectProductWithCompleteGraph( C ) ) then 
        Info( InfoGpd, 1, "Option 1 in HomsetCosetsGroupoidCoset" ); 
        fgpd := FamilyObj( C ); 
        rays := List( obs, o -> One(ogp) ); 
    else 
        Info( InfoGpd, 1, "Option 2 in HomsetCosetsGroupoidCoset" ); 
        fgpd := FamilyObj( Parent( C ) );  
        rays := List( C!.rayConjugators, r -> r^(-1) ); 
        if isrt then 
            pos := Position( obs, o1 ); 
        else 
            pos := Position( obs, o2 ); 
        fi; 
        if ( pos <> 1 ) then   ## not the root object 
            rpos := rays[pos]^(-1); 
            if isrt then 
                rays := List( [1..Length(obs)], j -> rays[j]*rpos ); 
            else
                rays := List( [1..Length(obs)], j -> rpos*rays[j] ); 
            fi; 
        fi; 
    fi; 
    if isrt then 
        gpcset := RightCoset( ogp, e![1] ); 
        fam := FamilyObj( [ fgpd, gpcset, obs, [o2], rays, "r" ] );
        hc := Objectify( NewType( fam, filter ), 
                         [ fgpd, gpcset, obs, [o2], rays, "r" ] );
    elif IsMyLeftCosetDefaultRep( cset ) then 
        gpcset := RightCoset( ogp, e![1]^(-1) ); 
        fam := FamilyObj( [ fgpd, gpcset, [o1], obs, rays, "l" ] );
        hc := Objectify( NewType( fam, filter ), 
                         [ fgpd, gpcset, [o1], obs, rays, "l" ] ); 
    fi; 
    SetIsHomsetCosets( hc, true ); 
    return hc;
end);

#############################################################################
########  Why not `IsIdenticalObj' in the following declarations ??? ########

############################################################################# 
## 
#M  RightCosetRepresentatives
## 
InstallMethod( RightCosetRepresentatives, "generic method for a subgroupoid",
    true, [ IsGroupoid and IsSinglePiece, IsGroupoid ], 0,
function( G, U )

    local  gg, obG, nobG, reps, c, o1, gc, rc, nrc, L, j, b, g, o;

    if not IsWideSubgroupoid( G, U ) then
        Error( "U not a wide subgroupoid of G" );
    fi;
    gg := G!.magma;
    obG := G!.objects;
    nobG := Length( obG );
    reps := [ ];
    for c in Pieces( U ) do
        o1 := c!.objects[1];
        gc := c!.magma;
        rc := RightCosets( gg, gc );
        nrc := Length( rc );
        L := ListWithIdenticalEntries( nrc*nobG, 0 );
        j := 0;
        for b in rc do
            g := Representative(b);
            for o in obG do
                j := j+1;
                L[j] := GroupoidElementNC( g, o1, o );
            od;
        od;
        Append( reps, L );
    od;
    return reps;
end );    

InstallMethod( RightCosetRepresentatives, "generic method for a subgroupoid",
    true, [ IsGroupoid, IsGroupoid ], 0,
function( G, U )

    local  pos, cG, cU, ncU, reps, i, filt, piece, m, j, subU;

    if not IsWideSubgroupoid( G, U ) then
        Error( "U not a wide subgroupoid of G" );
    fi;
    pos := PiecePositions( G, U );
    reps := [ ];
    cG := Pieces( G );
    cU := Pieces( U );
    ncU := Length( cU );
    for i in [1..Length(cG)] do
        filt := Filtered( [1..ncU], p -> ( pos[p] = i ) );
        piece := List( filt, p -> cU[p] );
        if ( Length(filt) = 1 ) then
            subU := piece[1];
        else
            subU := UnionOfPiecesNC( piece );
        fi;
        Append( reps, RightCosetRepresentatives( cG[i], subU ) );
    od;
    return reps;
end );

############################################################################# 
## 
#M  LeftCosetRepresentatives
## 
InstallMethod( LeftCosetRepresentatives, "generic method for a subgroupoid",
    true, [ IsGroupoid and IsSinglePiece, IsGroupoid ], 0,
function( G, U )

    local  gg, obG, nobG, reps, c, o1, gc, rc, nrc, L, j, b, g, o;

    if not IsWideSubgroupoid( G, U ) then
        Error( "U not a wide subgroupoid of G" );
    fi;
    gg := G!.magma;
    obG := G!.objects;
    nobG := Length( obG );
    reps := [ ];
    for c in Pieces( U ) do
        o1 := c!.objects[1];
        gc := c!.magma;
        rc := RightCosets( gg, gc );
        nrc := Length( rc );
        L := ListWithIdenticalEntries( nrc*nobG, 0 );
        j := 0;
        for b in rc do
            g := Representative(b)^(-1);
            for o in obG do
                j := j+1;
#?                L[j] := GroupoidElementNC( G, g, o, o1 );
                L[j] := GroupoidElementNC( g, o, o1 );
            od;
        od;
        Append( reps, L );
    od;
    return reps;
end );    

InstallMethod( LeftCosetRepresentatives, "generic method for a subgroupoid",
    true, [ IsGroupoid, IsGroupoid ], 0,
function( G, U )

    local  pos, cG, cU, ncU, reps, i, filt, piece, m, j, subU;

    if not IsWideSubgroupoid( G, U ) then
        Error( "U not a wide subgroupoid of G" );
    fi;
    pos := PiecePositions( G, U );
    reps := [ ];
    cG := Pieces( G );
    cU := Pieces( U );
    ncU := Length( cU );
    for i in [1..Length(cG)] do
        filt := Filtered( [1..ncU], p -> ( pos[p] = i ) );
        piece := List( filt, p -> cU[p] );
        if ( Length(filt) = 1 ) then
            subU := piece[1];
        else
            subU := UnionOfPiecesNC( piece );
        fi;
        Append( reps, LeftCosetRepresentatives( cG[i], subU ) );
    od;
    return reps;
end );

############################################################################# 
## 
#M  LeftCosetRepresentativesFromObject
## 
InstallMethod( LeftCosetRepresentativesFromObject, "for a subgroupoid",
    true, [ IsGroupoid and IsSinglePiece, IsGroupoid, IsScalar ], 0,
function( G, U, o )

    local  gg, obG, nobG, reps, c, o1, gc, rc, nrc, L, j, b, g;

    if not IsWideSubgroupoid( G, U ) then
        Error( "U not a wide subgroupoid of G" );
    fi;
    gg := G!.magma;
    obG := G!.objects;
    if not ( o in obG ) then
        Error( "chosen object not in the groupoid" );
    fi;
    reps := [ ];
    for c in Pieces( U ) do
        o1 := c!.objects[1];
        gc := c!.magma;
        rc := RightCosets( gg, gc );
        nrc := Length( rc );
        L := ListWithIdenticalEntries( nrc, 0 );
        j := 0;
        for b in rc do
            g := Representative(b)^(-1);
            j := j+1;
#?            L[j] := GroupoidElementNC( G, g, o, o1 );
            L[j] := GroupoidElementNC( g, o, o1 );
        od;
        Append( reps, L );
    od;
    return reps;
end );    

InstallMethod( LeftCosetRepresentativesFromObject, "for a subgroupoid",
    true, [ IsGroupoid, IsGroupoid, IsScalar ], 0,
function( G, U, o )

    local  pos, cG, co, i, cU, ncU, filt, piece, m, j, subU;

    if not IsWideSubgroupoid( G, U ) then
        Error( "U not a wide subgroupoid of G" );
    fi;
    pos := PiecePositions( G, U );
    cG := Pieces( G );
    co := PieceOfObject( G, o );
    if ( co = fail ) then
        Error( "chosen object not in the groupoid" );
    fi;
    i := Position( cG, co );
    cU := Pieces( U );
    ncU := Length( cU );
    filt := Filtered( [1..ncU], p -> ( pos[p] = i ) );
    piece := List( filt, p -> cU[p] );
    if ( Length(filt) = 1 ) then
        subU := piece[1];
    else
        subU := UnionOfPiecesNC( piece );
    fi;
    return LeftCosetRepresentativesFromObject( co, subU, o );
end );

############################################################################# 
## 
#M  DoubleCosetRepresentatives
## 
InstallMethod( DoubleCosetRepresentatives, "generic method for 2 subgroupoids",
    true, [ IsGroupoid and IsSinglePiece, IsGroupoid, IsGroupoid ], 0,
function( G, U, V )

    local  gg, obG, nobG, reps, cu, ou, gu, cv, ov, gv, dc, c, r;

    if not ( IsWideSubgroupoid( G, U ) and IsWideSubgroupoid( G, V ) ) then
        Error( "U,V not wide subgroupoids of G" );
    fi;
    gg := G!.magma;
    obG := G!.objects;
    nobG := Length( obG );
    reps := [ ];
    for cu in Pieces( U ) do
        ou := cu!.objects[1];
        gu := cu!.magma;
        for cv in Pieces( V ) do
            ov := cv!.objects[1];
            gv := cv!.magma;
            dc := DoubleCosets( gg, gu, gv );
            for c in dc do
                r := Representative( c );
#?                Add( reps, GroupoidElementNC( G,r,ou,ov ) );
                Add( reps, GroupoidElementNC( r, ou, ov ) );
            od;
        od;
    od;
    return reps;
end );    

##  still need to implement  DoubleCosetRepresentatives for G not connected

#############################################################################
##
#M  RightCosetsNC 
#M  LeftCosetsNC 
#M  DoubleCosetsNC 
##
##  right cosets refine costars and left cosets refine stars 
##
##  RightCosets(G,U) is normally defined only when U is wide in G. 
##  For computational purposes we shall use, for the more general case, 
##  RightCosets(F,U) where F is the FullSubgroupoid of G on Objects(U). 
##
InstallOtherMethod( RightCosetsNC, "for groupoids", true,
    [ IsGroupoid, IsGroupoid ], 0,
function( G, U )

    local  cosets, reps, nr, i, e;

    reps := RightCosetRepresentatives( G, U );
    nr := Length( reps );
    cosets := ListWithIdenticalEntries( nr, 0 );
    for i in [1..nr] do
        e := reps[i];
        cosets[i] := RightCoset( G, U, e );
    od;
    return cosets;
end);
    
InstallMethod( LeftCosetsNC, "for groupoids", true, 
    [ IsGroupoid, IsGroupoid ], 0,
function( G, U )

    local  cosets, reps, nr, i, e;

    reps := LeftCosetRepresentatives( G, U );
    nr := Length( reps );
    cosets := ListWithIdenticalEntries( nr, 0 );
    for i in [1..nr] do
        e := reps[i];
        cosets[i] := LeftCoset( G, U, e );
    od;
    return cosets;
end);
    
InstallOtherMethod( DoubleCosetsNC, "for groupoids", true,
    [ IsGroupoid, IsGroupoid, IsGroupoid ], 0,
function( G, U, V )

    local  cosets, reps, nr, i, e;

    reps := DoubleCosetRepresentatives( G, U, V );
    nr := Length( reps );
    cosets := ListWithIdenticalEntries( nr, 0 );
    for i in [1..nr] do
        e := reps[i];
        cosets[i] := DoubleCoset( U, e, V );
    od;
    return cosets;
end);
    
#############################################################################
##
##  GpdBuildManual()
##
GpdBuildManual:=function()
local gpd_path, gpd_main, gpd_files, gpd_bookname;
gpd_path:=Concatenation(
               GAPInfo.PackagesInfo.("gpd")[1].InstallationPath,"/doc/");
gpd_main := "manual.xml";
gpd_bookname:="Gpd";
MakeGAPDocDoc( gpd_path, gpd_main, [ ], gpd_bookname );;
end;

##############################################################################
##
#E  gpd.gi . . . . . . . . . . . . . . . . . . . . . . . . . . . . . ends here
##
