##############################################################################
##
#W  gpdhom.gi                 GAP4 package `gpd'                 Chris Wensley
#W                                                                & Emma Moore
##
#H  @(#)Id: "mag/gap/gpdhom.gi", v.1.05 2008/11/18 
##
##  This file contains generic methods for magmas
##
Revision.("gpd/gap/gpdhom_gi") :=
             "@(#)$Id: gpdhom.gi,v 1.05 2008/11/18 gap Exp $";

#############################################################################
##  Standard error messages

GROUPOID_MAPPING_CONSTRUCTORS := Concatenation(
    "The standard operations which construct a groupoid mapping are:\n",
    "1.  GroupoidHomomorphismFromSinglePiece( src, rng, hom, imobs, rays );\n",
    "2.  HomomorphismToSinglePiece( src, rng, list of [hom,imobs,rays]'s );\n",
    "3.  HomomorphismByUnion( src, rng, list of disjoint mappings );\n", 
    "4.  GroupoidHomomorphism( one of the previous parameter options );" );

#############################################################################
##
#F  GroupoidHomomorphism( <g1>,<g2>,<maps> )                union of mappings 
#F  GroupoidHomomorphism( <g1>,<g2>,<piece images> )       single piece range 
#F  GroupoidHomomorphism( <g1>,<g2>,<hom>,<imobs>,<rays> )   connected source 
##
InstallGlobalFunction( GroupoidHomomorphism, function( arg )

    local  nargs, id, rays;
    nargs := Length( arg );

    if ( ( nargs < 3 ) or not IsMagmaWithObjects( arg[1] ) or 
         ( nargs > 5 ) or not IsMagmaWithObjects( arg[2] ) ) then 
        Info( InfoGpd, 1, GROUPOID_MAPPING_CONSTRUCTORS );
        return fail;
    fi;
    # mwo, mwo, group homomorphism, object images, ray multipliers 
    if ( ( nargs = 5 ) and IsSinglePiece( arg[1] ) 
                       and IsGroupHomomorphism( arg[3] ) 
                       and IsHomogeneousList( arg[4] ) 
                       and IsHomogeneousList( arg[5] ) ) then 
        Info( InfoGpd, 2, "MagmaHomomorphismFromSinglePiece" );
        return GroupoidHomomorphismFromSinglePiece(
                   arg[1], arg[2], arg[3], arg[4], arg[5] ); 
    # mwo, mwo, group homomorphism, object images
    elif ( ( nargs = 4 ) and IsSinglePiece( arg[1] ) 
                       and IsGroupHomomorphism( arg[3] ) 
                       and IsHomogeneousList( arg[4] ) ) then 
        id := One( Range( arg[3] ) ); 
        rays := List( arg[4], o -> id ); 
        Info( InfoGpd, 2, "MagmaHomomorphismFromSinglePiece, adding id rays" );
        return GroupoidHomomorphismFromSinglePiece( 
                   arg[1], arg[2], arg[3], arg[4], rays ); 
    # mwo, mwo, list of mappings
    elif ( ( nargs = 3 ) and IsHomogeneousList( arg[3] ) 
            and IsHomomorphismToSinglePiece( arg[3][1] ) ) then
        Info( InfoGpd, 2, "HomomorphismByUnion" );
        return HomomorphismByUnion( arg[1], arg[2], arg[3] );
    # gpd, gpd, list of [hom, images of objects, rays] triples
    elif ( ( nargs = 3 ) and IsHomogeneousList( arg[3] ) 
           and IsList( arg[3][1] ) and IsList( arg[3][1][1] ) ) then 
        Info( InfoGpd, 2, "HomomorphismToSinglePiece" );
        return HomomorphismToSinglePiece( arg[1], arg[2], arg[3] );
    else
        Info( InfoGpd, 1, GROUPOID_MAPPING_CONSTRUCTORS );
        return fail;
    fi;
end );

#############################################################################
##
#M  IsomorphismPermGroupoid
##
InstallMethod( IsomorphismPermGroupoid, "for a connected groupoid", true,
    [ IsGroupoid and IsSinglePiece ], 0,
function( g1 )

    local  iso, obs, g, g2p, p, g2;

    obs := g1!.objects;
    g := g1!.magma;
    g2p := IsomorphismPermGroupoid( g );
    p := Image( g2p );
    g2 := SinglePieceMagmaWithObjects( obs, p );
    iso := GroupoidHomomorphismFromSinglePiece( g1, g2, obs, g2p );
    SetIsInjectiveWithObjects( iso, true );
    SetIsSurjectiveWithObjects( iso, true );
    return iso;
end );

InstallMethod( IsomorphismPermGroupoid, "generic method for a groupoid", 
    true, [ IsGroupoid ], 0,
function( g1 )

    local  isos, comp1, nc1, i, g2, iso;

    comp1 := Pieces( g1 );
    nc1 := Length( comp1 );
    isos := ListWithIdenticalEntries( nc1, 0 );
    for i in [1..nc1] do 
        isos[i] := IsomorphismPermGroupoid( comp1[i] );
    od;
    g2 := UnionOfPieces( List( isos, m -> Image(m) ) );
    iso := HomomorphismByUnion( g1, g2, isos );
    SetIsInjectiveWithObjects( iso, true );
    SetIsSurjectiveWithObjects( iso, true );
    return iso;
end );

#############################################################################
##
#M  InclusionMappingGroupoids
##
InstallMethod( InclusionMappingGroupoids, "for a subgroupoid of a groupoid", 
    true, [ IsGroupoid and IsSinglePiece, IsGroupoid ], 0,
function( gpd, sgpd )
    local  sobs, o1, c1, mappings, comps, m, inc, mor;

    if not IsSubgroupoid( gpd, sgpd ) then
        Error( "arg[2] is not a subgroupoid of arg[1]" );
    fi;
    if IsSinglePiece( sgpd ) then
        sobs := sgpd!.objects;
        if IsSinglePiece( gpd ) then
            inc := InclusionMappingGroups( gpd!.magma, sgpd!.magma );
            if ( inc = fail ) then
                Error( "magma mapping fails to include" );
            fi;
            return HomomorphismToSinglePieceNC(sgpd,gpd,[[sobs,inc]]);
        else
            o1 := sgpd!.objects[1];
            c1 := PieceOfObject( gpd, o1 );
            inc := InclusionMappingGroups( c1!.magma, sgpd!.magma );
            if ( inc = fail ) then
                Error( "magma mapping fails to include" );
            fi;
            mor := HomomorphismToSinglePieceNC(sgpd,c1,[[sobs,inc]]);
            return MagmaWithObjectsHomomorphism( sgpd, gpd, [ mor ] );
        fi;
    else
        Info( InfoGpd, 2, "InclusionMapping, source not connected" );
        mappings := List( Pieces( sgpd ),
                           c -> InclusionMappingGroupoids( gpd, c ) );
        return HomomorphismByUnion( sgpd, gpd, mappings );
    fi;
end );

InstallMethod( InclusionMappingGroupoids, "for a subgroupoid of a groupoid", 
    true, [ IsGroupoid, IsGroupoid ], 0,
function( gpd, sgpd )
    if not IsSubdomainWithObjects( gpd, sgpd ) then
        Error( "arg[2] is not a submagma of arg[1]" );
    fi;
    if not IsSinglePiece( sgpd ) then
        Error( "not yet implemented when arg[2] is not connected" );
    fi;
end );

#############################################################################
##
#M  RestrictionMappingGroupoids
##
InstallMethod( RestrictionMappingGroupoids, "for a groupoid mapping", true,
    [ IsGroupoidHomomorphism, IsGroupoid, IsGroupoid ], 0,
function( mor, U, V )

    local  src, rng, images, hom, imo, obsU, compU, lenU, mors, j, sj, osj, 
           pos, orj, rj, rhom;

    src := Source( mor );
    if not IsSinglePiece( src ) then
        Print( "not yet implemented for Source(mor) not connected\n" );
        return fail;
    fi;
    if not IsSinglePiece( V ) then
        Print( "not yet implemented for V not connected\n" );
        return fail;
    fi;
    rng := Range( mor );
    if not ( IsSubdomainWithObjects( src, U ) 
             and IsSubdomainWithObjects( rng, V ) ) then
        Error( "not a pair of submagmas" );
    fi;
    images := PieceImages( mor );
    imo := images[1][1];
    hom := images[1][2];
    obsU := ObjectList( U );
    compU := Pieces( U );
    lenU := Length( compU );
    mors := ListWithIdenticalEntries( lenU, 0 );
    for j in [1..lenU] do
        sj := compU[j];
        osj := sj!.objects;
        orj := List( osj, o -> imo[Position(obsU,o)] );
        rhom := RestrictionMappingGroups( hom, sj!.magma, rng!.magma );
        if ( rhom = fail ) then
            Error( "magma mapping fails to restrict" );
        fi;
        mors[j] := GroupoidHomomorphismFromSinglePiece( sj, rng, orj, rhom );
    od;
    return HomomorphismByUnion( U, V, mors );
end );

##  this one not checked
InstallMethod( RestrictionMappingGroupoids, "for a groupoid mapping", true,
    [ IsGroupoidHomomorphism, IsGroupoid, IsGroupoid ], 0,
function( mor, ssrc, srng )

    local  comp, len, rcomp, j, m;

    comp := PiecesOfMapping( mor );
    len := Length( comp );
    rcomp := ListWithIdenticalEntries( len, 0 );
    for j in [1..len] do
        rcomp[j] := RestrictionMappingGroupoids( comp[j], ssrc, srng );
    od;
    return HomomorphismByUnionNC( Source(mor), Range(mor), rcomp );
end );

#############################################################################
##
#M  InverseGeneralMapping 
##
InstallMethod( InverseGeneralMapping, "for a connected mapping", true,
    [  IsGroupoidHomomorphism and IsHomomorphismFromSinglePiece ], 0,
function( map )

    local  m1, m2, ob1, ob2, nob, imobs, one2, hom21, len, sc1, sc2, rays, inv;

    #?  (14/11/08)  is this special method really necessary ?? 
    if not (IsInjectiveWithObjects(map) and IsSurjectiveWithObjects(map)) then
        Error( "mapping not bijective on objects" );
    fi;
    m1 := Source( map );
    m2 := Range( map );
    ob1 := m1!.objects;
    ob2 := m2!.objects;
    nob := Length( ob1 );
    imobs := ImagesOfObjects( map );
    one2 := One( m2!.magma );
    if not ( LeftProdsOfRays(map) = List( ob2, o -> one2 ) ) then 
        Error( "not yet coded the general case" ); 
    fi;
    sc1 := ShallowCopy( ob1 );
    sc2 := ShallowCopy( imobs );
    SortParallel( sc2, sc1 );
    hom21 := InverseGeneralMapping( RootHomomorphism( map ) );
    #?  why is the following trest necessary ?? 
    SetIsGroupHomomorphism( hom21, true );
    SetIsTotal( hom21, true );
    SetRespectsMultiplication( hom21, true );
    SetIsInjective( hom21, true );
    SetIsSurjective( hom21, true );
    rays := List( ob2, o -> One( m1!.magma ) ); 
    inv := GroupoidHomomorphismFromSinglePieceNC( m2, m1, hom21, sc1, rays );
    SetIsInjectiveWithObjects( inv, true );
    SetIsSurjectiveWithObjects( inv, true );
    return inv;
end );

############################################################################# 
##
#M  GroupoidHomomorphismFromDiscreteToSinglePiece
##
## InstallMethod( GroupoidHomomorphismFromDiscreteToSinglePiece,
##     "for a list of groupoid mappings with common connected range", true,
##     [ IsGroupoid and IsDiscrete, IsGroupoid and IsSinglePiece, 
##       IsHomogeneousList ], 0,  
## function( src, rng, maps )


####################### functions from the 08.05 paper ####################### 

#############################################################################
##
#M  MappingPermObjectsImages . . . . . . for list of objects and their images 
##
InstallMethod( MappingPermObjectsImages, "for objects and their images", true, 
    [ IsList, IsList ], 0,
function( obs, ims )
    local  len, L; 
    len := Length( obs ); 
    if ( len <> Length(ims) ) then
        Error("<obs> and <ims> have different lengths");
    fi; 
    L := ShallowCopy( ims ); 
    Sort( L ); 
    if not ( L = obs ) then 
        return fail; 
    fi;
    L := List( obs, o -> Position( ims, o ) ); 
    return PermList( L ); 
end );

#############################################################################
##
#M  ObjectPermOfGroupoidHomomorphism . . . . . . . . . . .  for a groupoid mapping 
##
InstallMethod( ObjectPermOfGroupoidHomomorphism, "for objects and their images", 
    true, [ IsDefaultGroupoidHomomorphismRep ], 0,
function( map )
    local  obs, ims; 
    obs := Range( map )!.objects; 
    ims := ImagesOfObjects( map ); 
    if not ( Length(obs) = Length(ims) ) then 
        Info( InfoGpd, 1, "ObjectPermOfGroupoidHomomorphism set to <fail>" );
        return fail; 
    fi; 
    return MappingPermObjectsImages( obs, ims );
end ); 

#############################################################################
##
#M  Display
##
InstallMethod( Display, "for a mapping of connected groupoids", true,
    [ IsGroupoidHomomorphism and IsHomomorphismToSinglePiece ], 0,
function ( map )
    Print( "groupoid mapping: " );
    Print( "[ ", Source(map), " ] -> [ ", Range(map), " ]\n" ); 
    Print( "root homomorphism: ", 
           MappingGeneratorsImages( RootHomomorphism(map) ), "\n" ); 
    Print( "images of objects: ", ImagesOfObjects( map ), "\n" );
    Print( "   images of rays: ", LeftProdsOfRays( map ), "\n" ); 
end );

#############################################################################
##
#M  GroupoidHomomorphismFromSinglePieceNC  
#M  GroupoidHomomorphismFromSinglePiece 
##
InstallMethod( GroupoidHomomorphismFromSinglePieceNC,
    "generic method for a mapping of connected groupoids", true,
    [ IsGroupoid and IsSinglePiece, IsGroupoid and IsSinglePiece, 
      IsGroupHomomorphism, IsHomogeneousList, IsHomogeneousList ], 0,
function( src, rng, hom, oims, rims )

    local  fam, filter, map, ok, p;

    fam := FamilyObj( [ src, rng, hom, oims, rims ] );
    filter := IsDefaultGroupoidHomomorphismRep and IsGroupoidHomomorphism; 
    map := rec(); 
    ObjectifyWithAttributes( map, NewType( fam, filter ), 
        Source, src, 
        Range, rng, 
        RootHomomorphism, hom, 
        ImagesOfObjects, oims, 
        LeftProdsOfRays, rims,   #? ); 
        IsGeneralMappingWithObjects, true, 
        IsGroupoidHomomorphism, true, 
        IsHomomorphismToSinglePiece, true, 
        RespectsMultiplication, true ); 
    ok := IsInjectiveWithObjects( map ); 
    ok := IsSurjectiveWithObjects( map ); 
    p := ObjectPermOfGroupoidHomomorphism( map ); 
    SetIsHomomorphismFromSinglePiece( map, true ); 
    #?  (10/10/08) do we really need this ?? 
    SetPieceImages( map, [ hom, oims, rims ] );
    if ( src = rng ) then 
        SetIsEndoGeneralMapping( map, true ); 
        SetIsEndomorphismWithObjects( map, true ); 
        ##ok := IsAutomorphismWithObjects( map ); 
    fi; 
    return map; 
end );

InstallMethod( GroupoidHomomorphismFromSinglePiece,
    "generic method for a mapping of connected groupoids", true,
    [ IsGroupoid and IsSinglePiece, IsGroupoid and IsSinglePiece, 
      IsGroupHomomorphism, IsHomogeneousList, IsHomogeneousList ], 0,
function( src, rng, hom, oims, rims )

    local  ok; 
#? (10/10/08) insert some checks here!
    Info( InfoGpd, 1, "need to insert checks in this function" ); 
    return GroupoidHomomorphismFromSinglePieceNC( src, rng, hom, oims, rims ); 
end );

#############################################################################
##
#M  GroupoidHomomorphismByObjectImagesNC  
#M  GroupoidHomomorphismByObjectImages 
##
InstallMethod( GroupoidHomomorphismByObjectImagesNC , 
    "for a groupoid and a permutation of objects", true, 
    [ IsGroupoid and IsSinglePiece, IsGroupoid and IsSinglePiece, 
      IsHomogeneousList ], 0,
function( src, rng, oims )
    local  rgp, robs, hom, rims, pos, rays; 
    rgp := rng!.magma; 
    hom := InclusionMappingGroups( rgp, src!.magma ); 
    robs := rng!.objects; 
    if ( HasIsDirectProductWithCompleteGraph( rng ) 
         and IsDirectProductWithCompleteGraph( rng ) ) then 
        rims := List( oims, o -> One( rgp ) ); 
    else 
        rays := rng!.rayConjugators; 
        pos := List( oims, o -> Position( robs, o ) ); 
        rims := List( [1..Length(pos)], i -> rays[ pos[i] ] ); 
    fi; 
    return GroupoidHomomorphismFromSinglePieceNC( src, rng, hom, oims, rims ); 
end ); 

InstallMethod( GroupoidHomomorphismByObjectImages, 
    "for a groupoid and a permutation of objects", true, 
    [ IsGroupoid and IsSinglePiece, IsGroupoid and IsSinglePiece, 
      IsHomogeneousList ], 0,
function( src, rng, oims )
    local  sgp, rgp, cims, pos, rays; 
    sgp := src!.magma; 
    rgp := rng!.magma;   #? use ObjectGroup here ?? 
    if not IsSubgroup( rgp, sgp ) then 
        Error( "root group of source not <= that of range" ); 
    fi;  
    if not ( Length( src!.objects ) = Length( oims ) ) then 
        Error( "require image of each object in source" ); 
    fi; 
    cims := ShallowCopy( oims ); 
    Sort( cims ); 
    if not IsSubset( rng!.objects, cims ) then 
        Error( "images of objects not all in range groupoid" ); 
    fi; 
    return GroupoidHomomorphismByObjectImagesNC( src, rng, oims ); 
end ); 

#############################################################################
##
#M  GroupoidHomomorphismByGroupHomNC  
#M  GroupoidHomomorphismByGroupHom 
##
InstallMethod( GroupoidHomomorphismByGroupHomNC , 
    "for a groupoid and a homomorphism from the root group", true, 
    [ IsGroupoid and IsSinglePiece, IsGroupoid and IsSinglePiece, 
      IsGroupHomomorphism ], 0,
function( src, rng, hom )
    local  rgp, oims, rims, pos, rays; 
    rgp := rng!.magma; 
    oims := rng!.objects;
    if ( HasIsDirectProductWithCompleteGraph( rng ) 
         and IsDirectProductWithCompleteGraph( rng ) ) then 
        rims := List( oims, o -> One( rgp ) ); 
    else 
        rims := rng!.rayConjugators; 
    fi; 
    return GroupoidHomomorphismFromSinglePieceNC( src, rng, hom, oims, rims ); 
end ); 

InstallMethod( GroupoidHomomorphismByGroupHom, 
    "for a groupoid and a homomorphism from the root group", true, 
    [ IsGroupoid and IsSinglePiece, IsGroupoid and IsSinglePiece, 
      IsGroupHomomorphism ], 0,
function( src, rng, hom )  
    if not ( Source(hom) = src!.magma ) then 
        Error( "source of hom not the root group of src" ); 
    fi;  
    if not IsSubgroup( rng!.magma, Range(hom) ) then 
        #? (10/10/08) should use ObjectGroup here ?? 
        Error( "range of hom not a subgroup of the root group of rng" ); 
    fi; 
    if not ( Length( src!.objects ) = Length( rng!.objects ) ) then 
        Error( "src and rng do not have the same number of objects" ); 
    fi;  
    return GroupoidHomomorphismByGroupHomNC( src, rng, hom ); 
end ); 

#############################################################################
##
#M  GroupoidHomomorphismByRayProdsNC  
#M  GroupoidHomomorphismByRayProds 
##
InstallMethod( GroupoidHomomorphismByRayProdsNC , 
    "for a groupoid and a list of ray images", true, 
    [ IsGroupoid and IsSinglePiece, IsGroupoid and IsSinglePiece, 
      IsHomogeneousList ], 0,
function( src, rng, prods )
    local  rgp, oims, hom, pos, rays, rims; 
    rgp := rng!.magma; 
    oims := rng!.objects;
    if ( HasIsDirectProductWithCompleteGraph( rng ) 
         and IsDirectProductWithCompleteGraph( rng ) ) then 
        rims := prods; 
    else 
        rays := rng!.rayConjugators; 
        rims := List( [1..Length(oims)], i -> prods[i]*rays[i] ); 
    fi; 
    hom := InclusionMappingGroups( rgp, src!.magma ); 
    return GroupoidHomomorphismFromSinglePieceNC( src, rng, hom, oims, rims ); 
end ); 

InstallMethod( GroupoidHomomorphismByRayProds, 
    "for a groupoid and a list of images for the rays", true, 
    [ IsGroupoid and IsSinglePiece, IsGroupoid and IsSinglePiece, 
      IsHomogeneousList ], 0,
function( src, rng, prods )  
    local  rgp; 
    rgp := rng!.magma; 
    #? (10/10/08) is this correct ?? 
    if not ForAll( prods, p -> p in rgp ) then 
        Error( "ray products not all in the root group of rng" ); 
    fi; 
    if not ( prods[1] = One(rgp) ) then 
        Error( "first product not the identity element" ); 
    fi; 
    if not ( Length( src!.objects ) = Length( rng!.objects ) ) then 
        Error( "src and rng do not have the same number of objects" ); 
    fi;  
    return GroupoidHomomorphismByRayProdsNC( src, rng, prods ); 
end ); 

#############################################################################
##
#M  \=( <m1>, <m2> )  . . . . . . . . . . . equality of two groupoid mappings
##
InstallMethod( \=, "for 2 connected groupoid mappings", true,
    [ IsDefaultGroupoidHomomorphismRep, 
      IsDefaultGroupoidHomomorphismRep ], 0,
function( m1, m2 )
    return ( ( Source( m1 ) =  Source( m2 ) ) and
             ( Range( m1 ) = Range( m2 ) ) and 
             ( RootHomomorphism( m1 ) = RootHomomorphism( m2 ) ) and 
             ( ImagesOfObjects( m1 ) = ImagesOfObjects( m2 ) ) and 
             ( LeftProdsOfRays( m1 ) = LeftProdsOfRays( m1 ) ) );
end );

#############################################################################
##
#M  \*( <m1>, <m2> )  . . . . . . . . . . . product of two groupoid mappings
##
InstallMethod( \*, "for 2 connected magma mappings", true,
    [ IsDefaultGroupoidHomomorphismRep, IsDefaultGroupoidHomomorphismRep ], 0,
function( m1, m2 )

    local  s1, r1, s2, r2, ob1, nob1, ob2, io1, io2, ir1, ir2, 
           j, o, p, h1, h2, mgi1, im12, hom, oims, rims;

    Info( InfoGpd, 2, "method for * with IsDefaultGroupoidHomomorphismRep" );
    s1 := Source( m1 ); 
    r1 := Range( m1 );
    s2 := Source( m2 );
    r2 := Range( m2 );
    if not IsSubdomainWithObjects( s2, r1 ) then
        Error( "Range(m1) not a subgroupoid of Source(m2)" );
    fi; 
    ob1 := s1!.objects;
    nob1 := Length( ob1 );
    ob2 := s2!.objects; 
    io1 := ImagesOfObjects( m1 ); 
    io2 := ImagesOfObjects( m2 ); 
    oims := [1..nob1]; 
    for j in [1..nob1] do
        o := io1[j];
        p := Position( ob2, o );
        oims[j] := io2[p];
    od;
    h1 := RootHomomorphism( m1 ); 
    mgi1 := MappingGeneratorsImages( h1 ); 
    h2 := RootHomomorphism( m2 ); 
    im12 := List( mgi1[2], x -> Image( h2, x ) ); 
    hom := GroupHomomorphismByImages( Source(h1), Range(h2), mgi1[1], im12 ); 
    ir1 := LeftProdsOfRays( m1 ); 
    ir2 := LeftProdsOfRays( m2 ); 
    #? (10/10/08) this is incorrect, but will do for now! 
    rims := List( [1..nob1], i -> ir1[i]*ir2[i] ); 
    return GroupoidHomomorphismFromSinglePieceNC( s1, r2, hom, oims, rims );
end );

#############################################################################
##
#M  ImageElm( <map>, <e> )
##
InstallOtherMethod( ImageElm, "for a groupoid mapping between single pieces", 
    true, [IsGroupoidHomomorphism and IsHomomorphismFromSinglePiece, 
    IsGroupoidElement], 0,
function ( map, e )

    local  mwo1, mwo2, imo, obs1, hom, pt1, ph1, t2, h2, g2, prods;

    Info( InfoGpd, 3, "using ImageElm for IsHomomorphismFromSinglePiece" ); 
    #?  need to include some tests here ?? 
    mwo1 := Source( map ); 
    mwo2 := Range( map ); 
    imo := ImagesOfObjects( map ); 
    obs1 := mwo1!.objects; 
    pt1 := Position( obs1, e![2] ); 
    ph1 := Position( obs1, e![3] ); 
    t2 := imo[ pt1 ];
    h2 := imo[ ph1 ]; 
    hom := RootHomomorphism( map ); 
    g2 := Image( hom, e![1] ); 
    prods := LeftProdsOfRays( map ); 
    if ( HasIsDirectProductWithCompleteGraph( mwo1 ) 
         and IsDirectProductWithCompleteGraph( mwo1 ) ) then 
        g2 := prods[pt1]^(-1) * g2 * prods[ph1]; 
    else 
        Error( "this case not yet allowed for" ); 
    fi;
    return MultiplicativeElementWithObjects( mwo2, g2, t2, h2 );
end );

InstallOtherMethod( ImageElm, "for a groupoid mapping between single pieces", 
    true, [IsGroupoidHomomorphism, IsGroupoidElement], 0,
function ( map, e )

    local  G1, ce, C1, pe, Ge, obs1, hom, obs2, mag2, t2, h2, g2;

    #?  need to include some tests here ??
    G1 := Source( map );
    ce := PieceOfObject( G1, e![2] );
    C1 := Pieces( G1 );
    pe := Position( C1, ce );
    Ge := C1[pe];
    obs1 := Ge!.objects;
    hom := PieceImages( map )[pe][2];
    obs2 := PieceImages( map )[pe][1]; 
    #?  this is not the correct range magma ??
    mag2 := Range( map ); 
    t2 := obs2[ Position( obs1, e![2] ) ];
    h2 := obs2[ Position( obs1, e![3] ) ]; 
    g2 := Image( hom, e![1] );
    return MultiplicativeElementWithObjects( mag2, g2, t2, h2 );
end );


##############################################################################
##
#E  gpdhom.gi  . . . . . . . . . . . . . . . . . . . . . . . . . . . ends here
##
