############################################################################# 
## 
#W  mwo.gi                    GAP4 package `Gpd'                Chris Wensley 
##
#H  @(#)Id: "gpd/gap/mwo.gi", v.1.05 2008/11/21 
## 
##  This file contains the declarations of elements, magma, etc., and their 
##  families in the case of many objects.  So each algebraic structure comes 
##  with a set of objects, and each element $e$ has a tail object $te$ and a 
##  head object $he$, and multiplies partially if composable: on the left 
##  with those elements in its family which have head $te$, and on the right 
##  with those elements of its family which have tail $he$.  Non-composable 
##  elements return $fail$ when multiplied. 
##  
Revision.("gpd/gap/mwo_gi") :=
             "@(#)$Id: mwo.gi,v 1.05 2008/11/21 gap Exp $";


###########################  DOMAIN WITH OBJECTS  ########################### 

############################################################################# 
## 
#M  TypeOfDomainWithObjects( <dwo> ) 
##
InstallMethod( TypeOfDomainWithObjects, "for a list of domains with objects", 
    true, [ IsList ], 0, 

    function( pieces ) 
    local  type; 
    ## type:  1=gpd, 2=mon, 3=sgp, 4=mgm, 5=dom 
    type := 0; 
    if ForAll( pieces, p -> HasIsGroupoid(p) and IsGroupoid(p) ) then 
        type := 1; 
    elif ForAll( pieces, p -> 
            HasIsMonoidWithObjects(p) and IsMonoidWithObjects(p) ) then 
        type := 2; 
    elif ForAll( pieces, p -> 
            HasIsSemigroupWithObjects(p) and IsSemigroupWithObjects(p) ) then 
        type := 3; 
    elif ForAll( pieces, p -> 
        ( "IsMagmaWithObjects" in CategoriesOfObject(p) ) 
                              and IsMagmaWithObjects(p) ) then 
        type := 4; 
    elif ForAll( pieces, p -> 
        ( "IsDomainWithObjects" in CategoriesOfObject(p) )  
                              and IsDomainWithObjects(p) ) then 
        type := 5; 
    else 
        Info( InfoGpd, 2, "just an ordinary list?" ); 
    fi; 
    return type; 
end );



#####################  MULT ELTS WITH OBJECTS  ############################## 

############################################################################# 
## 
#M  MultiplicativeElementWithObjects( <mwo>, <elt>, <tail>, <head> ) 
#M  MultiplicativeElementWithObjectsNC( <mwo>, <elt>, <tail>, <head> ) 
##
InstallMethod( MultiplicativeElementWithObjectsNC, 
    "for magma with objects, element, tail and head", true,  
    [ IsMagmaWithObjects, IsMultiplicativeElement, IsScalar, IsScalar ], 
    0, 
    function( mwo, e, t, h ) 
    local  obs, elt, fam;
    fam := mwo!.eltsfam; 
    elt := Objectify( 
        NewType( fam, IsMultiplicativeElementWithObjects ), 
        [ e, t, h ] ); 
    return elt; 
end ); 

InstallMethod( MultiplicativeElementWithObjects, 
    "for general magma with objects, element, tail and head", true, 
    [ IsMagmaWithObjects, IsMultiplicativeElement, IsScalar, IsScalar ], 0, 
    function( mwo, e, t, h ) 
    local  piece, obs, fam, mag, pwo, pos, homset, pose; 
    if IsSinglePiece( mwo ) then 
        piece := mwo; 
    else 
        piece := PieceOfObject( mwo, t );
    fi;
    mag := piece!.magma; 
    if not ( e in mag ) then 
        Error( "<e> not in magma <mag>" ); 
    fi;
    obs := piece!.objects; 
    if not ( ( t in obs ) and ( h in obs ) ) then  
        Error( "<t> and <h> must be objects in <piece>" ); 
    fi;
    if not IsDirectProductWithCompleteGraph( piece ) then 
        if not IsBound( piece!.table ) then 
            TryNextMethod(); 
        fi;
        pos := [ Position( obs, t ), Position( obs, h ) ]; 
        if not HasMultiplicationTable( mag ) then 
            Error( "expecting magma defined by multiplication table" ); 
        fi; 
        homset := piece!.table[pos[1]][pos[2]]; 
        pose := Position( GeneratorsOfMagma( mag ), e ); 
        if not ( pose in homset ) then 
            Error( "(e : t -> h) not an element in <piece>" ); 
        fi;   
    fi; 
    return MultiplicativeElementWithObjectsNC( piece, e, t, h ); 
end );

#############################################################################
## 
#M  Arrowelt
#M  Arrowtail
#M  Arrowhead
##
InstallMethod( Arrowelt, "generic method for a magma with objects element", 
    true, [ IsMultiplicativeElementWithObjects ], 0, e -> e![1] ); 

InstallMethod( Arrowtail, "generic method for a magma with objects element", 
    true, [ IsMultiplicativeElementWithObjects ], 0, e -> e![2] ); 

InstallMethod( Arrowhead, "generic method for a magma with objects element", 
    true, [ IsMultiplicativeElementWithObjects ], 0, e -> e![3] ); 

#############################################################################
##
#M  PrintObj . . . . . . . . . . . . . . for elements in a magma with objects 
#M  ViewObj  . . . . . . . . . . . . . . for elements in a magma with objects 
##
InstallMethod( PrintObj, "for an element in a magma with objects",
    [ IsMultiplicativeElementWithObjects ],
function ( e )
    Print( "[", e![1], " : ", e![2], " -> ", e![3], "]" );
end );

InstallMethod( ViewObj, "for an element in a magma with objects",
    [ IsMultiplicativeElementWithObjects ], PrintObj );

#############################################################################
##
#M  \=( <e1>, <e2> ) . . . . . . equality of elements in a magma with objects
##
InstallMethod( \=, "for two multiplicative elements with objects", 
    IsIdenticalObj, [ IsMultiplicativeElementWithObjects, 
                      IsMultiplicativeElementWithObjects ], 0,
function( e1, e2 )
    return ForAll( [1..3], i -> ( e1![i] = e2![i] ) ); 
end );

############################################################################# 
## 
#M  \*( e1, e2 ) . . . . . . composition of elements in a magma with objects 
## 
InstallMethod( \*, "for two elements in a magma with objects", IsIdenticalObj,
    [IsMultiplicativeElementWithObjects, IsMultiplicativeElementWithObjects], 
    0, 
    function( e1, e2 ) 
    local  prod; 
    ## elements are composable? 
    if ( ( e1![3] = e2![2] ) and 
         ( FamilyObj(e1)!.mwo = FamilyObj(e2)!.mwo ) ) then 
        return MultiplicativeElementWithObjectsNC( 
            FamilyObj(e1)!.mwo, e1![1]*e2![1], e1![2], e2![3] ); 
    else 
        return fail; 
    fi;  
end );

#############################################################################
##
#M  Order( <e> )  . . . . . . . . . . . . . . . . . . . of a groupoid element
##
InstallOtherMethod( Order, "for a multiplicative element with objects", true,
    [ IsMultiplicativeElementWithObjects ], 0,
function( e )
    local  ord;
    if not ( e![2] = e![3] ) then
        Error( "tail of e <> head of e" );
    fi;
    return Order( e![1] );
end );



################################  MAGMAS  ################################### 

############################################################################# 
## 
#F  MagmaWithObjects( <mag>, <obs> ) 
##
InstallGlobalFunction( MagmaWithObjects, function( arg ) 

    local  obs, mag;
    # list of objects and a magma 
    if ( ( Length( arg ) = 2 ) and IsMagma( arg[1] ) 
            and IsCollection( arg[2] ) and IsScalar( arg[2][1] ) ) then 
        mag := arg[1]; 
        obs := arg[2]; 
        if ( HasIsGeneratorsOfMagmaWithInverses( mag ) 
            and IsGeneratorsOfMagmaWithInverses( mag ) 
            and HasIsAssociative( mag ) and IsAssociative( mag ) ) then 
            Info( InfoGpd, 1, "SinglePieceGroupoid:-" );
            return SinglePieceGroupoid( mag, obs ); 
        elif ( HasIsMonoid( mag ) and IsMonoid( mag ) ) then 
            Info( InfoGpd, 1, "SinglePieceMonoidWithObjects:-" ); 
            return SinglePieceMonoidWithObjects( mag, obs ); 
        elif ( HasIsSemigroup( mag ) and IsSemigroup( mag ) ) then 
            Info( InfoGpd, 1, "SinglePieceSemigroupWithObjects:-" ); 
            return SinglePieceSemigroupWithObjects( mag, obs ); 
        else  ## it's just a magma 
            Info( InfoGpd, 1, "SinglePieceMagmaWithObjects:-" );
            return SinglePieceMagmaWithObjects( mag, obs ); 
        fi; 
    else 
        Error( "Current usage: MagmaWithObjects( <mag>, <obs> )" ); 
    fi; 
end ); 

############################################################################# 
## 
#M  SinglePieceMagmaWithObjects( <mag>, <obs> ) . . . . for magma and objects 
##
InstallMethod( SinglePieceMagmaWithObjects, 
    "for magma, objects", true, [ IsMagma, IsCollection ], 0, 
    function( mag, obs ) 
    local  cf, one, r, gens, mwo, fmwo, cmwo, isa, isc; 
    cf := CollectionsFamily( FamilyObj( mag ) ); 
    r := obs[1];  ## root object  
    isa := IsAssociative( mag ); 
    isc := IsCommutative( mag ); 
#?  how to make the next command work ?? 
##  ObjectifyWithAttributes( mwo, NewType( cf, IsMWOSinglePieceRep ), 
##         rec( objects := obs, magma := mag, eltsfam := cf ), 
##         IsAssociative, isa, IsCommutative, isc ); 
    mwo:= Objectify( NewType(  cf, IsMWOSinglePieceRep ), 
           rec( objects := obs, magma := mag, eltsfam := cf ) ); 
#?  is this eltsfam needed any more ?? 
#?  associative magmas are semigroups, so omit next line ?? 
    SetIsAssociative( mwo, IsAssociative( mag ) ); 
    SetIsCommutative( mwo, IsCommutative( mag ) ); 
    SetIsFinite( mwo, IsFinite( mag ) ); 
    SetIsSinglePieceDomain( mwo, true ); 
    SetIsDirectProductWithCompleteGraphDomain( mwo, true ); 
    fmwo := FamilyObj( mwo ); 
    fmwo!.mwo := mwo;
    Info( InfoGpd, 2, "fmwo!.mwo = ", mwo ); 
    if ( HasOne( mag ) and not ( One( mag ) = fail ) ) then 
        one := One( mag ); 
        gens := Concatenation( 
            List( GeneratorsOfMagmaWithOne( mag ),  
                       g -> MultiplicativeElementWithObjectsNC(mwo,g,r,r) ), 
            List( obs, o -> MultiplicativeElementWithObjectsNC(mwo,one,r,o))); 
        SetGeneratorsOfMagmaWithObjectsAndOnes( mwo, gens ); 
    fi;
    return mwo; 
end ); 

#############################################################################
##
#M  \=( <m1>, <m2> )  . . . . . . . test if two magmas with objects are equal
##
InstallMethod( \=, "for magmas with objects", IsIdenticalObj,
    [ IsMagmaWithObjects, IsMagmaWithObjects ], 
function ( m1, m2 ) 
    local  i, c1, c2;
    if ( IsSinglePiece(m1) and IsSinglePiece(m2) ) then 
        return ( ( m1!.objects=m2!.objects ) and ( m1!.magma=m2!.magma ) ); 
    elif ( IsSinglePiece(m1) or IsSinglePiece(m2) ) then 
        return false; 
    else 
        c1 := m1!.pieces; 
        c2 := m2!.pieces; 
        if not ( Length(m1) = Length(m2) ) then 
            return false; 
        else 
            for i in [1..Length(c1)] do 
                if ( c1[i] <> c2[i] ) then 
                    return false; 
                fi; 
            od; 
            return true; 
        fi; 
    fi;
end );

############################################################################# 
## 
#M  ViewObj( <mwo> ) . . . . . . . . . . . . . . . view a magma with objects 
#M  PrintObj( <mwo> ) . . . . . . . . . . . . . . print a magma with objects 
##

##  InstallMethod( ViewObj, "for a domain with objects", true, 
##      [ IsDomainWithObjects ], 0, PrintObj );  

InstallMethod( ViewObj, "for a single piece magma with objects", true, 
    [ IsSinglePiece ], 0,   
    function( mwo )
    local  type; 
    type := TypeOfDomainWithObjects( [ mwo ] ); 
    if ( type = 1 ) then 
        Print( "#I  should be using special groupoid method!\n" ); 
    elif ( type = 2 ) then 
        Print( "monoid with objects :-\n" ); 
    elif ( type = 3 ) then 
        Print( "semigroup with objects :-\n" ); 
    elif ( type = 4 ) then 
        Print( "magma with objects :-\n" ); 
    else 
        Print( "not yet implemented for general domains with objects\n" ); 
    fi; 
    Print( "    magma = ", mwo!.magma, "\n" ); 
    Print( "  objects = ", mwo!.objects, "\n" ); 
end );

InstallMethod( PrintObj, "for a single piece magma with objects", true, 
    [ IsSinglePiece ], 0,   
    function( mwo )
    local  type; 
    type := TypeOfDomainWithObjects( [ mwo ] ); 
    if ( type = 1 ) then 
        Print( "#I  should be using special groupoid method!\n" ); 
    elif ( type = 2 ) then 
        Print( "monoid with objects :-\n" ); 
    elif ( type = 3 ) then 
        Print( "semigroup with objects :-\n" ); 
    elif ( type = 4 ) then 
        Print( "magma with objects :-\n" ); 
    else 
        Print( "not yet implemented for general domains with objects\n" ); 
    fi; 
    Print( "    magma = ", mwo!.magma, "\n" ); 
    Print( "  objects = ", mwo!.objects, "\n" ); 
end );

InstallMethod( ViewObj, "for more than one piece", true, 
    [ IsDomainWithObjects and IsPiecesRep ], 10,   
    function( dwo )
    local  i, pieces, np, type; 
    pieces := Pieces( dwo ); 
    np := Length( pieces ); 
    type := TypeOfDomainWithObjects( Pieces( dwo ) ); 
    if (type=1) then Print( "groupoid" );  
      elif (type=2) then Print( "monoid with objects" ); 
      elif (type=3) then Print( "semigroup with objects" ); 
      elif (type=4) then Print( "magma with objects" ); 
      elif (type=5) then Print( "domain with objects" ); 
      elif (type=0) then Error( "invalid domain with objects" ); 
    fi;
    Print( " having ", np, " pieces :-\n" ); 
    for i in [1..np] do 
        Print( pieces[i], "\n" ); 
    od; 
end ); 

InstallMethod( PrintObj, "for more than one piece", true, 
    [ IsDomainWithObjects and IsPiecesRep ], 10,   
    function( dwo )
    local  i, pieces, np; 
    pieces := Pieces( dwo ); 
    np := Length( pieces ); 
    Print( "domain with objects having ", np, " pieces :-\n" ); 
    for i in [1..np] do 
        Print( pieces[i], "\n" ); 
    od; 
end ); 

##############################################################################
##
#M  Display( <mwo> ) . . . . . . . . . . . . . .  display a magma with objects
##
InstallMethod( Display, "for a mwo", [ IsMagmaWithObjects ],
function( mwo )
    
    local  comp, c, i, m, len;

    if IsSinglePiece( mwo ) then 
        if IsDirectProductWithCompleteGraph( mwo ) then 
            Print( "Single constituent magma with objects: " );
            if HasName( mwo ) then
                Print( mwo );
            fi;
            Print( "\n" ); 
            Print( "  objects: ", mwo!.objects, "\n" );
            m := mwo!.magma;
            Print( "    magma: " );
            if HasName( m ) then
                Print( m, " = <", GeneratorsOfMagma( m ), ">\n" );
            else
                Print( m, "\n" );
            fi;
        else
            TryNextMethod(); 
        fi; 
    else
        comp := Pieces( mwo );
        len := Length( comp );
        Print( "Magma with objects with ", len, " constituents:\n" );
        for i in [1..len] do
            c := comp[i];
            if IsDirectProductWithCompleteGraph( c ) then 
                Print( "< objects: ", c!.objects, "\n" );
                m := c!.magma;
                Print( "    magma: " );
                if HasName( m ) then
                    Print( m, " = <", GeneratorsOfMagma( m ), "> >\n" );
                else
                    Print( m, " >\n" );
                fi;
            else 
                TryNextMethod(); 
            fi; 
        od;
    fi;
end );

#############################################################################
##
#M  GeneratorsOfMagmaWithObjects( <mag> )  . . . . . . . of a connected magma
##
InstallMethod( GeneratorsOfMagmaWithObjects, 
    "for a single piece magma with objects", true, [ IsSinglePiece ], 0,
function( mwo )

    local  obs, nobs, o1, m, mgens, id, gens1, gens2, g, o;

    #?  (21/11/08)  need to check that this function does what it should ?? 
    obs := mwo!.objects;
    nobs := Length( obs );
    o1 := obs[1];
    m := mwo!.magma; 
    mgens := GeneratorsOfMagma( m ); 
    if ( HasOne(m) and not (One(m) = fail) ) then 
        id := One( m ); 
        if IsGroup( m ) then 
            gens1 := List( mgens, 
                g -> MultiplicativeElementWithObjects( mwo, g, o1, o1 ) );
            gens2 := List( obs{[2..nobs]}, 
                o -> MultiplicativeElementWithObjects( mwo, id, o1, o ) ); 
        else 
            gens1 := List( mgens, 
                g -> GroupoidElementNC( g, o1, o1 ) );
            gens2 := List( obs{[2..nobs]}, 
                o -> GroupoidElementNC( id, o1, o ) ); 
        fi;
        return Immutable( Concatenation( gens1, gens2 ) ); 
    else  
    ## add in all possible generators? 
        gens1 := [ ]; 
        for g in mgens do 
            for o in obs do 
                Append( gens1, [MultiplicativeElementWithObjects(mwo,g,o1,o)]);
            od; 
        od;
        return gens1; 
    fi; 
end );

InstallMethod( GeneratorsOfMagmaWithObjects, "for more than one constituent",
    true, [ IsMagmaWithObjects ], 0,
function( mwo )
    local  L; 
    L := Concatenation( List( Pieces( mwo ), 
                              p -> GeneratorsOfMagmaWithObjects( p ) ) ); 
    return L; 
end );



########################  MORE THAN ONE CONSTITUENT  ######################## 

#############################################################################
##
#M  Pieces . . . . . . . . . . connected components of a domain with objects 
##
InstallMethod( Pieces, "for a single piece domain with objects",
    true, [ IsSinglePieceDomain ], 0,
function( dwo )
    return [ dwo ];
end );

#############################################################################
##
#M  ObjectList . . . . . . . . . . . . . . . . . . . for a magma with objects
##
InstallMethod( ObjectList, "for a magma with objects",
    true, [ IsMagmaWithObjects ], 0,
function( mwo )

    local  obs; 

    if ( HasIsSinglePiece( mwo ) and IsSinglePiece( mwo ) ) then
        return mwo!.objects;
    else
        obs := Concatenation( 
            List( Pieces( mwo ), c -> c!.objects ) );
        if not IsDuplicateFree( obs ) then
            Error( "same object in more than one constituent" );
        fi;
        Sort( obs );
        return obs;
    fi;
end );

#############################################################################
##
#M  UnionOfPiecesNC . . . . . . . for a list of connected magmas with objects
#M  UnionOfPieces
##
InstallMethod( UnionOfPiecesNC, "method for magmas with objects",
    true, [ IsList, IsInt ], 0,
function( comps, type )

    local  len, pieces, L, fam, filter, mwo, i, obs, par;

    ## type:  1=gpd, 2=mon, 3=sgp, 4=mgm, 5=dom 

    ## order pieces by first object
    len := Length( comps ); 
    obs := List( comps, g -> g!.objects[1] );
    L := [1..len];
    SortParallel( obs, L );
    if ( L = [1..len] ) then 
        pieces := comps; 
    else 
        Info( InfoGpd, 2, "reordering pieces by first object" ); 
        pieces := List( L, i -> comps[i] );
    fi; 
    fam := FamilyObj( [ pieces ] ); 
    if ( type = 1 ) then 
        filter := IsPiecesRep and IsMagmaWithObjectsAndInverses 
                                    and IsAssociative; 
    elif ( type = 2 ) then 
        filter := IsPiecesRep and IsMagmaWithObjectsAndOnes 
                                    and IsAssociative; 
    elif ( type = 3 ) then 
        filter := IsPiecesRep and IsMagmaWithObjects and IsAssociative; 
    elif ( type = 4 ) then 
        filter := IsPiecesRep; 
    else 
        Error( "union of unstructured domains not yet implemented" ); 
    fi; 
    #?  should use ObjectifyWithAttributes here ??
    mwo := Objectify( NewType( fam, filter ), rec () );
    SetIsSinglePieceDomain( mwo, false ); 
    SetPieces( mwo, pieces ); 
    if HasParent( pieces[1] ) then 
        par := Ancestor( pieces[1] ); 
        if ForAll( pieces, c -> ( Ancestor(c)=par ) ) then 
            SetParent( mwo, par ); 
        fi; 
    fi; 
    if ( type = 1 ) then 
        if ForAll( pieces, p -> HasIsPermGroupoid(p) and IsPermGroupoid(p) )
             then SetIsPermGroupoid( mwo, true ); 
        elif ForAll( pieces, p -> HasIsPcGroupoid(p) and IsPcGroupoid(p) )
             then SetIsPcGroupoid( mwo, true ); 
        elif ForAll( pieces, p -> HasIsFpGroupoid(p) and IsFpGroupoid(p) )
             then SetIsFpGroupoid( mwo, true ); 
        fi; 
    fi; 
    return mwo; 
end );

InstallMethod( UnionOfPieces, "generic method for magmas with objects",
    true, [ IsList ], 0,
function( parts )

    local  npa, part, pieces, p, nco, obs, obp, i, gi, nobj, type;

    npa := Length( parts );
    pieces := [ ]; 
    if ( Length( parts ) = 1 ) then 
        Error( "only one part supplied" ); 
    fi; 
    for part in parts do 
        if not IsDomainWithObjects( part ) then
            Info( InfoGpd, 1, "part ", part, "not an mwo" );
            return fail;
        fi;
        if ( HasIsSinglePiece(part) and IsSinglePiece(part) ) then
            Add( pieces, part );
        else
            Append( pieces, Pieces( part ) );
        fi;
    od; 
    obs := [ ]; 
    for p in pieces do 
        obp := ObjectList( p );
        if ( Intersection( obs, obp ) <> [ ] ) then
            Info( InfoGpd, 1, "constituents must have disjoint object sets" );
            return fail;
        fi;
        obs := Union( obs, obp ); 
    od;
    type := TypeOfDomainWithObjects( pieces ); 
    Info( InfoGpd, 2, "union has type ", type ); 
    return UnionOfPiecesNC( pieces, type );
end );

#############################################################################
##
#M  DomainWithSingleObject
##
InstallMethod( DomainWithSingleObject, "generic method for domain, object",
    true, [ IsDomain, IsObject ], 0,
function( dom, obj )
    local  o; 
    if ( IsList( obj ) and ( Length(obj) = 1 ) ) then
        Info( InfoGpd, 2, "object given as a singleton list" );
        o := obj[1]; 
    else
        o := obj;
    fi; 
    if not IsScalar( o ) then 
        Error( "<obj> not a scalar or singleton list" ); 
    fi; 
    if ( HasIsAssociative( dom ) and IsAssociative( dom ) 
         and ( "IsMagmaWithInverses" in CategoriesOfObject( dom ) ) 
         and IsMagmaWithInverses( dom ) ) then 
        return SinglePieceGroupoid( dom, [o] ); 
    elif ( HasIsMonoid( dom ) and IsMonoid( dom ) ) then 
        return SinglePieceMonoidWithObjects( dom, [o] ); 
    elif ( HasIsSemigroup( dom ) and IsSemigroup( dom ) ) then 
        return SinglePieceSemigroupWithObjects( dom, [o] ); 
    elif ( ( "IsMagma" in CategoriesOfObject(dom) ) and IsMagma(dom) ) then 
        return SinglePieceMagmaWithObjects( dom, [o] ); 
    else 
        Error( "unstructured domains with objects not yet implemented" ); 
    fi; 
end );

############################################################################# 
## 
#M  PieceOfObject
## 
InstallMethod( PieceOfObject, "generic method for magma with objects", 
    true, [ IsDomainWithObjects, IsScalar ], 0,
function( dwo, obj )

    local  pieces, p, objp;

    if IsSinglePiece( dwo ) then
        if not ( obj in dwo!.objects ) then
            Error( "<obj> not an object of <dwo>" );
        else
            return dwo;
        fi;
    elif not ( obj in ObjectList( dwo ) ) then
        Info( InfoGpd, 1, "<obj> not an object of <dwo>" );
        return fail;
    fi;
    pieces := Pieces( dwo );
    for p in pieces do
        objp := p!.objects;
        if ( obj in objp ) then
            return p;
        fi;
    od;
    Info( InfoGpd, 1, "it appears that <obj> is not an object in <dwo>" );
    return fail;
end );

############################################################################# 
## 
#M  PieceNrOfObject
## 
InstallMethod( PieceNrOfObject, "generic method for domain with objects",
    true, [ IsDomainWithObjects, IsScalar ], 0,
function( dwo, obj )

    local  pieces, i, objp, np;
    pieces := Pieces( dwo );
    for i in [1..Length( pieces )] do
        objp := pieces[i]!.objects;
        if ( obj in objp ) then
            return i;
        fi;
    od;
    Info( InfoGpd, 1, "it appears that <obj> is not an object in <dwo>" );
    return fail;
end );

#############################################################################
##
#M  IsDiscreteDomain 
##
InstallMethod( IsDiscreteDomain, "for a magma with objects", true,
    [ IsDomainWithObjects ], 0,
function( dwo )

    local  p;
    for p in Pieces( dwo ) do
        if not ( Length( p!.objects ) = 1 ) then
            return false;
        fi;
    od;
    return true;
end );



#################################  SUBDOMAINS  ############################## 

#############################################################################
##
#F  IsSubdomainWithObjects( <M>, <U> )
##
InstallMethod( IsSubdomainWithObjects, "for two domains with objects", true, 
    [ IsDomainWithObjects, IsDomainWithObjects ], 0, 
    function( D, U )

    local  compU, obj, p, ok; 

    #?  (21/11/08)  does any of this make sence ?? 
    #?  does a subdomeain of a groupoid have to be a groupoid ?? 
    ok := false; 
    if not ( IsMagmaWithObjects(D) and IsMagmaWithObjects(U) ) then
        Error( "not yet implemented for unstructured domains" ); 
    fi;
    if ( HasParentAttr(U) and ( ParentAttr(U) = D ) ) then 
        return true;
    fi;
    if not IsSubset( ObjectList(D), ObjectList(U) ) then
        return false;
    fi; 
    if ( IsSinglePiece(D) and IsSinglePiece(U) ) then
        obj := U!.objects[1];
        if ( IsGroupoid(D) and IsGroupoid(U) ) then 
            #?  (21/11/08)  needs more testing here ?? 
            ok := IsSubgroup( ObjectGroup(D,obj), ObjectGroup(U,obj) ); 
##      elif ( IsMonoid(D) and IsMonoid(U) ) then 
##          ok := ( IsSubsemigroup( D!.magma, U!.magma ) 
##                  and IsMonoid( D!.magma ) ); 
##      elif ( IsSemigroup(D) and IsSemigroup(U) ) then 
##          ok := IsSubsemigroup( D!.magma, U!.magma ); 
        else 
            ok := IsSubset( GeneratorsOfMagma( D!.magma ), 
                            GeneratorsOfMagma( U!.magma ) ); 
        fi;
    elif IsSinglePiece(U) then
        obj := U!.objects[1];
        p := PieceOfObject( D, obj ); 
        ok := IsSubdomainWithObjects( p, U );
    else
        compU := Pieces(U);
        ok := ForAll( compU, p -> IsSubdomainWithObjects( D, p ) );
    fi;
    if ( ok and not HasParentAttr( U ) ) then 
        SetParent( U, D ); 
    fi; 
    return ok; 
end );




################################  SEMIGROUPS  ############################### 

############################################################################# 
## 
#F  SemigroupWithObjects( <sgp>, <obs> ) 
##
InstallGlobalFunction( SemigroupWithObjects, function( arg ) 

    local  obs, sgp;
    # list of objects and a semigroup 
    if ( ( Length( arg ) = 2 ) and IsSemigroup( arg[1] ) 
            and IsCollection( arg[2] ) and IsScalar( arg[2][1] ) ) then 
        sgp := arg[1]; 
        obs := arg[2]; 
        if ( HasIsGeneratorsOfMagmaWithInverses( sgp ) 
            and IsGeneratorsOfMagmaWithInverses( sgp ) 
            and HasIsAssociative( sgp ) and IsAssociative( sgp ) ) then  
            return SinglePieceGroupoid( sgp, obs ); 
        elif ( HasIsMonoid( sgp ) and IsMonoid( sgp ) ) then 
            return SinglePieceMonoidWithObjects( sgp, obs ); 
        else  ## it's just a semigroup
            return SinglePieceSemigroupWithObjects( sgp, obs ); 
        fi; 
    else 
        Error( "Current usage: SemigroupWithObjects( <sgp>, <obs> )" ); 
    fi; 
end ); 

############################################################################# 
## 
#M  SinglePieceSemigroupWithObjects( <sgp>, <obs> ) . for semigroup, objects
##
InstallMethod( SinglePieceSemigroupWithObjects, 
    "for objects, semigroup", true, [ IsSemigroup, IsCollection ], 0, 
    function( sgp, obs ) 

    local  cf, one, r, gens, swo, fswo, cswo, isa, isc; 

    cf := CollectionsFamily( FamilyObj( sgp ) ); 
    r := obs[1];  ## root object  
    isa := IsAssociative( sgp ); 
    isc := IsCommutative( sgp ); 
    swo:= Objectify( NewType(  cf, IsMWOSinglePieceRep ), 
           rec( objects := obs, magma := sgp, eltsfam := cf ) ); 
    SetIsAssociative( swo, IsAssociative( sgp ) ); 
    SetIsCommutative( swo, IsCommutative( sgp ) ); 
    SetIsFinite( swo, IsFinite( sgp ) ); 
    SetIsSinglePieceDomain( swo, true ); 
    SetIsDirectProductWithCompleteGraphDomain( swo, true ); 
    fswo := FamilyObj( swo ); 
    fswo!.mwo := swo; 
    if ( HasOne( sgp ) and not ( One( sgp ) = fail ) ) then 
        one := One( sgp ); 
        gens := Concatenation( 
            List( GeneratorsOfMagmaWithOne( sgp ),  
                       g -> MultiplicativeElementWithObjectsNC(sgp,g,r,r) ), 
            List( obs, o -> MultiplicativeElementWithObjectsNC(sgp,one,r,o))); 
        SetGeneratorsOfMagmaWithObjectsAndOnes( swo, gens ); 
    fi;
    return swo; 
end ); 




##################################  MONOIDS  ################################ 

############################################################################# 
## 
#F  MonoidWithObjects( <mon>, <obs> ) 
##
InstallGlobalFunction( MonoidWithObjects, function( arg ) 

    local  obs, mon;
    # list of objects and a monoid 
    if ( ( Length( arg ) = 2 ) and IsMonoid( arg[1] ) 
            and IsCollection( arg[2] ) and IsScalar( arg[2][1] ) ) then 
        mon := arg[1]; 
        obs := arg[2]; 
        if ( HasIsGeneratorsOfMagmaWithInverses( mon ) 
            and IsGeneratorsOfMagmaWithInverses( mon ) 
            and HasIsAssociative( mon ) and IsAssociative( mon ) ) then  
            return SinglePieceGroupoid( mon, obs ); 
        else  ## it's just a monoid
            return SinglePieceMonoidWithObjects( mon, obs ); 
        fi; 
    else 
        Error( "Current usage: MonoidWithObjects( <mon>, <obs> )" ); 
    fi; 
end ); 

############################################################################# 
## 
#M  SinglePieceMonoidWithObjects( <mon>, <sgp> ) . . . for semigroup, objects 
##
InstallMethod( SinglePieceMonoidWithObjects, 
    "for objects, monoid", true, [ IsMonoid, IsCollection ], 0, 
    function( mon, obs ) 

    local  cf, one, r, gens, mwo, fmwo, cmwo, isa, isc; 

    cf := CollectionsFamily( FamilyObj( mon ) ); 
    r := obs[1];  ## root object  
    isa := IsAssociative( mon ); 
    isc := IsCommutative( mon ); 
    mwo:= Objectify( NewType(  cf, IsMWOSinglePieceRep 
                               and IsMagmaWithObjectsAndOnes ), 
           rec( objects := obs, magma := mon, eltsfam := cf ) ); 
    SetIsAssociative( mwo, IsAssociative( mon ) ); 
    SetIsCommutative( mwo, IsCommutative( mon ) ); 
    SetIsFinite( mwo, IsFinite( mon ) ); 
    SetIsSinglePieceDomain( mwo, true ); 
    SetIsDirectProductWithCompleteGraphDomain( mwo, true ); 
    fmwo := FamilyObj( mwo ); 
    fmwo!.mwo := mwo; 
    if ( HasOne( mon ) and not ( One( mon ) = fail ) ) then 
        one := One( mon ); 
        gens := Concatenation( 
            List( GeneratorsOfMagmaWithOne( mon ),  
                       g -> MultiplicativeElementWithObjectsNC(mwo,g,r,r) ), 
            List( obs, o -> MultiplicativeElementWithObjectsNC(mwo,one,r,o))); 
        SetGeneratorsOfMagmaWithObjectsAndOnes( mwo, gens ); 
    fi;
    return mwo; 
end ); 




##################################  GROUPS  ################################# 
##  A *group with objects* is a magma with objects where the vertex magmas 
##  are groups, and so is a *groupoid* - see file gpd.gi.




#################################  SUBMAGMAS  ############################### 

############################################################################# 
## 
#M  IsSubmagmaWithObjectsGeneratingTable( <mag>, <A> ) 
##
InstallMethod( IsSubmagmaWithObjectsGeneratingTable, "for magma and array", 
    true, [ IsMagma, IsList ], 0, 
    function( mag, A ) 
    local n, s, a, b, e;
    n := Length( A );    ## number of objects 
    s := Size( mag );  
    e := [1..s]; 
    for a in A do 
        if not ( IsList(a) and ( Length(a) = n ) ) then 
            return false; 
        fi; 
        for b in a do 
            if not IsSubset( e, b ) then 
                return false; 
            fi; 
        od; 
    od; 
    return true; 
end ); 

############################################################################# 
## 
#M  SubmagmaWithObjectsElementsTable( <mag>, <A> ) 
##
InstallMethod( SubmagmaWithObjectsElementsTable, "for magma and array", 
    true, [ IsMagma, IsList ], 0, 
    function( mag, A ) 
    local t, done, B, C, T, n, s, i, j, k, Lij, Lik, Ljk, a, b, ab; 
    if not IsSubmagmaWithObjectsGeneratingTable( mag, A ) then 
        Error( "array A is not a generating table for a submagma over mag" ); 
    fi; 
    T := MultiplicationTable( mag ); 
    s := Size( mag ); 
    n := Length( A ); 
    done := false; 
    C := StructuralCopy( A ); 
    t := 0; 
    while not done do 
        t := t+1; 
        B := StructuralCopy( C );   ## C stores all the elements 
        for i in [1..n] do 
            for j in [1..n] do 
                Lij := C[i][j]; 
                for k in [1..n] do 
                    Ljk := C[j][k]; 
                    for a in Lij do 
                        for b in Ljk do 
                            ab := T[a][b]; 
                            C[i][k] := Union( C[i][k], [ab] ); 
                        od; 
                    od; 
                od; 
            od; 
        od; 
        done := ( B = C ); 
    od;
    return C;        
end ); 

############################################################################# 
## 
#M  SubmagmaWithObjectsByElementsTable( <mwo>, <A> ) . . for mwo and elements 
##
InstallMethod( SubmagmaWithObjectsByElementsTable, 
    "for objects, magma and table of generating elements", true, 
    [ IsSinglePiece, IsList ], 0, 
    function( mwo, A ) 
    local  C, mag, obs, cf, isa, isc, swo; 
    obs := mwo!.objects; 
    mag := mwo!.magma; 
    C := SubmagmaWithObjectsElementsTable( mag, A ); 
    cf := mwo!.eltsfam; 
    isa := IsAssociative( mag ); 
    isc := IsCommutative( mag ); 
    swo:= Objectify( NewType(  cf, IsSubmagmaWithObjectsTableRep ), 
           rec( objects := obs, magma := mag, eltsfam := cf, table := C ) ); 
    SetParentAttr( swo, mwo );
    SetIsAssociative( swo, IsAssociative( mag ) ); 
    SetIsCommutative( swo, IsCommutative( mag ) ); 
    SetIsFinite( swo, IsFinite( mag ) ); 
    SetIsSinglePieceDomain( swo, true ); 
    #? (13/10/08)  still need to set GeneratorsOfMagma ?? 
    return swo; 
end ); 

############################################################################# 
## 
#M  PrintObj( <swo> ) . . . . .  for submagma with objects and elements table
##

InstallMethod( PrintObj, "for a submagma with objects and elements table", 
    true, [ IsSubmagmaWithObjectsTableRep ], 0, 
    function( M ) 
    Print( "submagma with objects:-\n" ); 
    Print( "objects = ", M!.objects, "\n" );
    Print( "  magma = ", M!.magma, "\n" );  
    Print( "  table = ", M!.table, "\n" ); 
end );


#################################  UTILITIES  ############################### 

############################################################################# 
## 
#M  Ancestor 
## 
InstallMethod( Ancestor, "method for a domain", 
    true, [ IsDomainWithObjects ], 0,
function( dom ) 
    local  found, d; 

    found := false; 
    d := dom; 
    while HasParent( d ) do 
        found := true; 
        d := Parent( d ); 
    od; 
    if not found then 
        return fail; 
    else 
        return d; 
    fi; 
end );


#############################################################################
##
#M  Iterator( <mwo> ) . . . .  iterator for a single piece magma with objects
##
InstallMethod( Iterator, "for a single piece magma with objects", 
    [ IsSinglePiece ], 
function( mwo )
    return IteratorByFunctions( rec( 
        IsDoneIterator := function( iter )
            return ( IsDoneIterator( iter!.magmaIterator ) 
                     and ( iter!.tpos = iter!.len )
                     and ( iter!.hpos = iter!.len ) );
            end, 
        NextIterator := function( iter )
            if ( iter!.tpos = 0 ) then
                iter!.melt := NextIterator( iter!.magmaIterator );
                iter!.tpos := 1;
                iter!.hpos := 1;
            elif ((iter!.tpos = iter!.len) and (iter!.hpos = iter!.len)) then 
                iter!.melt := NextIterator( iter!.magmaIterator );
                iter!.tpos := 1;
                iter!.hpos := 1;
            elif ( iter!.hpos = iter!.len ) then
                iter!.hpos := 1;
                iter!.tpos := iter!.tpos + 1;
            else 
                iter!.hpos := iter!.hpos + 1;
            fi;
            if ( HasIsDirectProductWithCompleteGraph( mwo ) 
                 and IsDirectProductWithCompleteGraph( mwo ) ) then 
               return MultiplicativeElementWithObjects( FamilyObj(mwo)!.mwo, 
                   iter!.melt, iter!.obs[iter!.tpos], iter!.obs[iter!.hpos] );
            else 
               return fail;
            fi; 
            end, 
        ShallowCopy := iter -> 
            rec( magmaIterator := ShallowCopy( iter!.magmaIterator ), 
                 melt := iter!.melt,
                 obs := iter!.obs,
                 len := iter!.len,
                 tpos := iter!.tpos,
                 hpos := iter!.hpos ),
        magmaIterator := Iterator( mwo!.magma ), 
        melt := 0, 
        obs := mwo!.objects,
        len := Length( mwo!.objects ),
        tpos := 0,
        hpos := 0 ) );
end );

InstallMethod( Iterator, "generic method for a magma with objects", 
    [ IsMagmaWithObjects ], 
function( mwo )
    return IteratorByFunctions( rec( 
        IsDoneIterator := function( iter )
            return ( IsDoneIterator( iter!.mwoIterator ) 
                     and ( iter!.cpos = iter!.len ) );
            end, 
        NextIterator := function( iter )
            if IsDoneIterator( iter!.mwoIterator ) then 
                iter!.cpos := iter!.cpos + 1;
                iter!.mwoIterator := 
                    Iterator( iter!.pieces[iter!.cpos] );
            fi;
            return NextIterator( iter!.mwoIterator );
            end,
        ShallowCopy := iter -> 
            rec( piecess := iter!.pieces,
                 len := iter!.len,
                 mwoIterator := ShallowCopy( iter!.mwoIterator ),
                 cpos := iter!.cpos ),
        pieces := Pieces( mwo ),
        len := Length( Pieces( mwo ) ),
        mwoIterator := Iterator( Pieces( mwo )[1] ),
        cpos := 1 ) );
end );

#############################################################################
##
#E  mwo.gi  . . . . . . . . . . . . . . . . . . . . . . . . . . . . ends here
##  
